from __future__ import annotations

from ftag.labels import LabelContainer

Flavours = LabelContainer.from_yaml(
    exclude_categories=[
        "single-btag-extended",
        "single-btag-extended-ghost",
    ]
)
Extended_Flavours = LabelContainer.from_yaml(
    exclude_categories=[
        "single-btag",
        "single-btag-ghost",
    ]
)



from __future__ import annotations

import re
from collections.abc import Iterable, Iterator
from dataclasses import dataclass
from pathlib import Path

import yaml

from ftag.cuts import Cuts


def remove_suffix(string: str, suffix: str) -> str:
    """Remove the suffix from a string.

    Parameters
    ----------
    string : str
        String from which the suffix is to be removed
    suffix : str
        Suffix to remove

    Returns
    -------
    str
        String with the suffix removed
    """
    if string.endswith(suffix):
        return string[: -len(suffix)]
    return string


@dataclass(frozen=True)
class Label:
    """Dataclass to hold info about one flavour/label.

    Attributes
    ----------
    name : str
        Name of the flavour/label
    label : str
        Plot label for this flavour/label
    cuts : Cuts
        Cuts of this flavour/label
    colour : str
        Colour of this flavour/label
    category : str
        Flavour category, to which this flavour/label belongs
    """

    name: str
    label: str
    cuts: Cuts
    colour: str
    category: str
    _px: str | None = None

    @property
    def px(self) -> str:
        return self._px or f"p{remove_suffix(self.name, 'jets')}"

    @property
    def eff_str(self) -> str:
        return self.label.replace("jets", "jet") + " efficiency"

    @property
    def rej_str(self) -> str:
        return self.label.replace("jets", "jet") + " rejection"

    @property
    def frac_str(self) -> str:
        return "f" + remove_suffix(self.name, "jets")

    def __str__(self) -> str:
        return self.name

    def __lt__(self, other) -> bool:
        return self.name < other.name


@dataclass
class LabelContainer:
    """Label container that holds multiple labels.

    Attributes
    ----------
    labels : dict[str, Label]
        Dict with the labels this container will hold
    """

    labels: dict[str, Label]

    def __iter__(self) -> Iterator:
        yield from self.labels.values()

    def __getitem__(self, key) -> Label:
        if isinstance(key, Label):
            key = key.name
        try:
            return self.labels[key]
        except KeyError as e:
            raise KeyError(f"Label '{key}' not found") from e

    def __len__(self) -> int:
        return len(self.labels.keys())

    def __getattr__(self, name) -> Label:
        return self[name]

    def __contains__(self, label: str | Label) -> bool:
        if isinstance(label, Label):
            label = label.name
        return label in self.labels

    def __eq__(self, other) -> bool:
        if isinstance(other, LabelContainer):
            return self.labels == other.labels
        if isinstance(other, list) and all(isinstance(f, str) for f in other):
            return {f.name for f in self} == set(other)
        return False

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}({', '.join([f.name for f in self])})"

    @property
    def categories(self) -> list[str]:
        return list(dict.fromkeys(f.category for f in self))

    def by_category(self, category: str) -> LabelContainer:
        f = LabelContainer({k: v for k, v in self.labels.items() if v.category == category})
        if not f.labels:
            raise KeyError(f"No labels with category '{category}' found")
        return f

    def from_cuts(self, cuts: list | Cuts) -> Label:
        if isinstance(cuts, list):
            cuts = Cuts.from_list(cuts)
        for label in self:
            if label.cuts == cuts:
                return label
        raise KeyError(f"Label with {cuts} not found")

    @classmethod
    def from_yaml(
        cls,
        yaml_path: Path | None = None,
        include_categories: Iterable[str] | None = None,
        exclude_categories: Iterable[str] | None = None,
    ) -> LabelContainer:
        if yaml_path is None:
            yaml_path = Path(__file__).parent / "flavours.yaml"
        with open(yaml_path) as f:
            config = yaml.safe_load(f)

        # Filter for categories to include
        if include_categories is not None:
            include_categories = set(include_categories)
            config = [f for f in config if f.get("category") in include_categories]

        # Filter for categories to exclude
        if exclude_categories is not None:
            exclude_categories = set(exclude_categories)
            config = [f for f in config if f.get("category") not in exclude_categories]

        if not config:
            raise KeyError("No labels left after category filtering.")

        # sanity checks
        cuts = [Cuts.from_list(f["cuts"]) for f in config]
        if duplicates := [c for c in cuts if cuts.count(c) > 1]:
            raise ValueError(f"Duplicate label definitions detected: {duplicates}")
        names = [f["name"] for f in config]
        if duplicates := [n for n in names if names.count(n) > 1]:
            raise ValueError(f"Duplicate label names detected: {duplicates}")

        labels = {f["name"]: Label(cuts=Cuts.from_list(f.pop("cuts")), **f) for f in config}
        return cls(labels)

    @classmethod
    def from_list(cls, labels: list[Label]) -> LabelContainer:
        return cls({f.name: f for f in labels})

    def backgrounds(self, signal: Label, only_signals: bool = True) -> LabelContainer:
        bkg = [f for f in self if f.category == signal.category and f != signal]
        if not only_signals:
            bkg = [f for f in bkg if f.name not in {"ujets", "qcd"}]
        if len(bkg) == 0:
            raise TypeError(
                "No background flavour could be found in the flavours for signal "
                f"flavour {signal.name}"
            )
        return LabelContainer.from_list(bkg)

    def cut_variables(self) -> list[str]:
        """Return all variable names appearing in any Label cuts.

        Returns
        -------
        list[str]
            Unique variable names used across all cuts in all labels.
        """
        vars_found: set[str] = set()
        var_regex = re.compile(r"[A-Za-z_]\w*")

        for label in self:
            for cut in label.cuts:
                tokens = var_regex.findall(str(cut))
                # filter out boolean keywords
                vars_found.update(tokens)

        return list(vars_found)



from __future__ import annotations

from dataclasses import dataclass

from ftag.cuts import Cuts


@dataclass(frozen=True)
class Region:
    """Dataclass to store info about a region.

    Attributes
    ----------
    name : str
        Name of the region
    cuts : Cuts
        Cuts of this region
    """

    name: str
    cuts: Cuts

    def __str__(self) -> str:
        return self.name

    def __lt__(self, other):
        return self.name < other.name

    def __eq__(self, other):
        return self.name == other.name



from __future__ import annotations

import argparse
from pathlib import Path


def valid_path(string: str) -> Path:
    """Check if the given string is a valid path.

    Parameters
    ----------
    string : str
        Input path as string

    Returns
    -------
    Path
        Output path as Path object

    Raises
    ------
    FileNotFoundError
        If the given input path as string doesn't exist
    """
    if (path := Path(string)).is_file():
        return path
    raise FileNotFoundError(string)


class HelpFormatter(  # noqa: D101
    argparse.RawTextHelpFormatter, argparse.ArgumentDefaultsHelpFormatter
): ...



"""Calculate tagger working points."""

from __future__ import annotations

import argparse
import sys
from pathlib import Path
from typing import TYPE_CHECKING

import numpy as np
import yaml

from ftag import Flavours
from ftag.cli_utils import HelpFormatter
from ftag.cuts import Cuts
from ftag.hdf5 import H5Reader
from ftag.utils import get_discriminant

if TYPE_CHECKING:  # pragma: no cover
    from collections.abc import Sequence

    from ftag.labels import Label, LabelContainer


def parse_args(args: Sequence[str]) -> argparse.Namespace:
    """Parse the input arguments into a Namespace.

    Parameters
    ----------
    args : Sequence[str]
        Sequence of string inputs to the script

    Returns
    -------
    argparse.Namespace
        Namespace with the parsed arguments

    Raises
    ------
    ValueError
        When both --effs and --disc_cuts are provided or
        When neither --effs nor --disc_cuts are provided or
        When the number of fraction values is not conistent or
        When the sum of fraction values for a tagger is not equal to one
    """
    # Define the pre-parser which checks the --category
    pre_parser = argparse.ArgumentParser(add_help=False)
    pre_parser.add_argument(
        "-c",
        "--category",
        default="single-btag",
        type=str,
        help="Label category to use for the working point calculation",
    )

    pre_parser.add_argument(
        "-s",
        "--signal",
        default="bjets",
        type=str,
        help="Signal flavour which is to be used",
    )

    # Parse only --category/--signal and ignore for now all other args
    pre_args, remaining_argv = pre_parser.parse_known_args(args=args)

    # Create the "real" parser
    parser = argparse.ArgumentParser(
        description=__doc__,
        formatter_class=HelpFormatter,
    )

    # Add --category/--signal so the help is correctly shown
    parser.add_argument(
        "-c",
        "--category",
        default="single-btag",
        type=str,
        help="Label category to use for the working point calculation",
    )
    parser.add_argument(
        "-s",
        "--signal",
        default="bjets",
        type=str,
        help="Signal flavour which is to be used",
    )

    # Check which label category was chosen and load the corresponding flavours
    flavours = Flavours.by_category(pre_args.category)

    # Build the fraction value arguments for all classes (besides signal)
    for flav in flavours:
        # Skip signal
        if flav.name == pre_args.signal:
            continue

        # Built fraction values for all background classes
        parser.add_argument(
            f"--{flav.frac_str}",
            nargs="+",
            required=True,
            type=float,
            help=f"{flav.frac_str} value(s) for each tagger",
        )

    # # Adding the other arguments
    parser.add_argument(
        "--ttbar",
        required=True,
        type=Path,
        help="Path to ttbar sample (supports globbing)",
    )
    parser.add_argument(
        "--zprime",
        required=False,
        type=Path,
        help="Path to zprime (supports globbing). WPs from ttbar will be reused for zprime",
    )
    parser.add_argument(
        "-t",
        "--tagger",
        nargs="+",
        required=True,
        type=str,
        help="tagger name(s)",
    )
    parser.add_argument(
        "-e",
        "--effs",
        nargs="+",
        type=float,
        help="Efficiency working point(s). If -r is specified, values should be 1/efficiency",
    )
    parser.add_argument(
        "-r",
        "--rejection",
        default=None,
        help="Use rejection of specified background class to determine working points",
    )
    parser.add_argument(
        "-d",
        "--disc_cuts",
        nargs="+",
        type=float,
        help="D_x value(s) to calculate efficiency at",
    )
    parser.add_argument(
        "-n",
        "--num_jets",
        default=1_000_000,
        type=int,
        help="Use this many jets (post selection)",
    )
    parser.add_argument(
        "--ttbar_cuts",
        nargs="+",
        default=["pt > 20e3"],
        type=list,
        help="Selection to apply to ttbar (|eta| < 2.5 is always applied)",
    )
    parser.add_argument(
        "--zprime_cuts",
        nargs="+",
        default=["pt > 250e3"],
        type=list,
        help="Selection to apply to zprime (|eta| < 2.5 is always applied)",
    )
    parser.add_argument(
        "-o",
        "--outfile",
        type=Path,
        help="Save results to yaml instead of printing",
    )

    # Final parse of all arguments
    parsed_args = parser.parse_args(remaining_argv)

    # Define the signal as an instance of Flavours
    parsed_args.signal = Flavours[parsed_args.signal]

    # Check that only --effs or --disc_cuts is given
    if parsed_args.effs and parsed_args.disc_cuts:
        raise ValueError("Cannot specify both --effs and --disc_cuts")
    if not parsed_args.effs and not parsed_args.disc_cuts:
        raise ValueError("Must specify either --effs or --disc_cuts")

    # Check that all fraction values have the same length
    for flav in flavours:
        if flav.name != parsed_args.signal.name and len(getattr(parsed_args, flav.frac_str)) != len(
            parsed_args.tagger
        ):
            raise ValueError(f"Number of {flav.frac_str} values must match number of taggers")

    # Check that all fraction value combinations add up to one
    for tagger_idx in range(len(parsed_args.tagger)):
        fraction_value_sum = 0
        for flav in flavours:
            if flav.name != parsed_args.signal.name:
                fraction_value_sum += getattr(parsed_args, flav.frac_str)[tagger_idx]

        # Round the value to take machine precision into account
        fraction_value_sum = np.round(fraction_value_sum, 8)

        # Check it's equal to one
        if fraction_value_sum != 1:
            raise ValueError(
                "Sum of the fraction values must be one! You gave "
                f"{fraction_value_sum} for tagger {parsed_args.tagger[tagger_idx]}"
            )
    return parsed_args


def get_fxs_from_args(args: argparse.Namespace, flavours: LabelContainer) -> list:
    """Get the fraction values for each tagger from the argparsed inputs.

    Parameters
    ----------
    args : argparse.Namespace
        Input arguments parsed by the argparser
    flavours : LabelContainer
        LabelContainer instance of the labels that are used

    Returns
    -------
    list
        List of dicts with the fraction values. Each dict is for one tagger.
    """
    # Init the fraction_dict dict
    fraction_dict = {}

    # Add the fraction values to the dict
    for flav in flavours:
        if flav.name != args.signal.name:
            fraction_dict[flav.frac_str] = vars(args)[flav.frac_str]

    return [{k: v[i] for k, v in fraction_dict.items()} for i in range(len(args.tagger))]


def get_eff_rej(
    jets: np.ndarray,
    disc: np.ndarray,
    wp: float,
    flavours: LabelContainer,
) -> dict:
    """Calculate the efficiency/rejection for each flavour.

    Parameters
    ----------
    jets : np.ndarray
        Loaded jets
    disc : np.ndarray
        Discriminant values of the jets
    wp : float
        Working point that is used
    flavours : LabelContainer
        LabelContainer instance of the flavours used

    Returns
    -------
    dict
        Dict with the efficiency/rejection values for each flavour
    """
    # Init an out dict
    out: dict[str, dict] = {"eff": {}, "rej": {}}

    # Loop over the flavours
    for flav in flavours:
        # Calculate discriminant values and efficiencies/rejections
        flav_disc = disc[flav.cuts(jets).idx]
        eff = sum(flav_disc > wp) / len(flav_disc)
        out["eff"][flav.name] = float(f"{eff:.3g}")
        out["rej"][flav.name] = float(f"{1 / eff:.3g}")

    return out


def get_rej_eff_at_disc(
    jets: np.ndarray,
    tagger: str,
    signal: Label,
    disc_cuts: list,
    flavours: LabelContainer,
    fraction_values: dict,
) -> dict:
    """Calculate the efficiency/rejection at a certain discriminant values.

    Parameters
    ----------
    jets : np.ndarray
        Loaded jets used
    tagger : str
        Name of the tagger
    signal : Label
        Label instance of the signal flavour
    disc_cuts : list
        List of discriminant cut values for which the efficiency/rejection is calculated
    flavours : LabelContainer
        LabelContainer instance of the flavours that are used
    fraction_values : dict
        Dict with the fraction values to use

    Returns
    -------
    dict
        Dict with the discriminant cut values and their respective efficiencies/rejections
    """
    # Calculate discriminants
    disc = get_discriminant(
        jets=jets,
        tagger=tagger,
        signal=signal,
        flavours=flavours,
        fraction_values=fraction_values,
    )

    # Init out dict
    ref_eff_dict: dict[str, dict] = {}

    # Loop over the disc cut values
    for dcut in disc_cuts:
        ref_eff_dict[str(dcut)] = {"eff": {}, "rej": {}}

        # Loop over the flavours
        for flav in flavours:
            e_discs = disc[flav.cuts(jets).idx]
            eff = sum(e_discs > dcut) / len(e_discs)
            ref_eff_dict[str(dcut)]["eff"][str(flav)] = float(f"{eff:.3g}")
            ref_eff_dict[str(dcut)]["rej"][str(flav)] = 1 / float(f"{eff:.3g}")

    return ref_eff_dict


def setup_common_parts(
    args: argparse.Namespace,
) -> tuple[np.ndarray, np.ndarray | None, LabelContainer]:
    """Load the jets from the files and setup the taggers.

    Parameters
    ----------
    args : argparse.Namespace
        Input arguments from the argparser

    Returns
    -------
    tuple[np.ndarray, np.ndarray | None, LabelContainer]
        Outputs the ttbar jets, the zprime jets (if wanted, else None), and the flavours used.
    """
    # Get the used flavours
    flavours = Flavours.by_category(args.category)

    # Get the cuts for the samples
    default_cuts = Cuts.from_list(["eta > -2.5", "eta < 2.5"])
    ttbar_cuts = Cuts.from_list(args.ttbar_cuts) + default_cuts
    zprime_cuts = Cuts.from_list(args.zprime_cuts) + default_cuts

    # Prepare the loading of the jets
    all_vars = list(set(sum((flav.cuts.variables for flav in flavours), [])))
    reader = H5Reader(args.ttbar)
    jet_vars = reader.dtypes()["jets"].names

    # Create for all taggers the fraction values
    for tagger in args.tagger:
        all_vars += [
            f"{tagger}_{flav.px}" for flav in flavours if (f"{tagger}_{flav.px}" in jet_vars)
        ]

    # Load ttbar jets
    ttbar_jets = reader.load({"jets": all_vars}, args.num_jets, cuts=ttbar_cuts)["jets"]
    zprime_jets = None

    # Load zprime jets if needed
    if args.zprime:
        zprime_reader = H5Reader(args.zprime)
        zprime_jets = zprime_reader.load({"jets": all_vars}, args.num_jets, cuts=zprime_cuts)[
            "jets"
        ]

    else:
        zprime_jets = None

    return ttbar_jets, zprime_jets, flavours


def get_working_points(args: argparse.Namespace) -> dict | None:
    """Calculate the working points.

    Parameters
    ----------
    args : argparse.Namespace
        Input arguments from the argparser

    Returns
    -------
    dict | None
        Dict with the working points. If args.outfile is given, the function returns None and
        stored the resulting dict in a yaml file in args.outfile.
    """
    # Load the jets and flavours and get the fraction values
    ttbar_jets, zprime_jets, flavours = setup_common_parts(args=args)
    fraction_values = get_fxs_from_args(args=args, flavours=flavours)

    # Init an out dict
    out = {}

    # Loop over taggers
    for i, tagger in enumerate(args.tagger):
        # Calculate discriminant
        out[tagger] = {"signal": str(args.signal), **fraction_values[i]}
        disc = get_discriminant(
            jets=ttbar_jets,
            tagger=tagger,
            signal=args.signal,
            flavours=flavours,
            fraction_values=fraction_values[i],
        )

        # Loop over efficiency working points
        for eff in args.effs:
            d = out[tagger][f"{eff:.0f}"] = {}

            # Set the working point
            wp_flavour = args.signal
            if args.rejection:
                eff = 100 / eff  # noqa: PLW2901
                wp_flavour = args.rejection

            # Calculate the discriminant value of the working point
            wp_disc = disc[flavours[wp_flavour].cuts(ttbar_jets).idx]
            wp = d["cut_value"] = round(float(np.percentile(wp_disc, 100 - eff)), 3)

            # Calculate efficiency and rejection for each flavour
            d["ttbar"] = get_eff_rej(
                jets=ttbar_jets,
                disc=disc,
                wp=wp,
                flavours=flavours,
            )

            # calculate for zprime
            if args.zprime:
                zprime_disc = get_discriminant(
                    jets=zprime_jets,
                    tagger=tagger,
                    signal=args.signal,
                    flavours=flavours,
                    fraction_values=fraction_values[i],
                )
                d["zprime"] = get_eff_rej(
                    jets=zprime_jets,
                    disc=zprime_disc,
                    wp=wp,
                    flavours=flavours,
                )

    if args.outfile:
        with open(args.outfile, "w") as f:
            yaml.dump(out, f, sort_keys=False)
            return None

    else:
        return out


def get_efficiencies(args: argparse.Namespace) -> dict | None:
    """Calculate the efficiencies for the given jets.

    Parameters
    ----------
    args : argparse.Namespace
        Input arguments from the argparser

    Returns
    -------
    dict | None
        Dict with the efficiencies. If args.outfile is given, the function returns None and
        stored the resulting dict in a yaml file in args.outfile.
    """
    # Load the jets and flavours and get the fraction values
    ttbar_jets, zprime_jets, flavours = setup_common_parts(args=args)
    fraction_values = get_fxs_from_args(args=args, flavours=flavours)

    # Init an out dict
    out = {}

    # Loop over the taggers
    for i, tagger in enumerate(args.tagger):
        out[tagger] = {"signal": str(args.signal), **fraction_values[i]}

        out[tagger]["ttbar"] = get_rej_eff_at_disc(
            jets=ttbar_jets,
            tagger=tagger,
            signal=args.signal,
            disc_cuts=args.disc_cuts,
            flavours=flavours,
            fraction_values=fraction_values[i],
        )
        if args.zprime:
            out[tagger]["zprime"] = get_rej_eff_at_disc(
                jets=zprime_jets,
                tagger=tagger,
                signal=args.signal,
                disc_cuts=args.disc_cuts,
                flavours=flavours,
                fraction_values=fraction_values[i],
            )

    if args.outfile:
        with open(args.outfile, "w") as f:
            yaml.dump(out, f, sort_keys=False)
            return None
    else:
        return out


def main(args: Sequence[str]) -> dict | None:
    """Main function to run working point calculation.

    Parameters
    ----------
    args : Sequence[str]
        Input arguments

    Returns
    -------
    dict | None
        The output dict with the calculated values. When --outfile
        was given, the return value is None
    """
    parsed_args = parse_args(args=args)

    if parsed_args.effs:
        out = get_working_points(args=parsed_args)

    elif parsed_args.disc_cuts:
        out = get_efficiencies(args=parsed_args)

    if out:
        print(yaml.dump(out, sort_keys=False))
        return out

    return None


if __name__ == "__main__":  # pragma: no cover
    main(args=sys.argv[1:])



from __future__ import annotations

import glob
from dataclasses import dataclass
from pathlib import Path

from ftag.labels import remove_suffix
from ftag.vds import create_virtual_file


@dataclass(frozen=True)
class Sample:
    """Dataclass which holds info about a specific sample.

    Attributes
    ----------
    pattern: Path | str | tuple[Path | str, ...]
        Filepattern for the h5 files
    ntuple_dir: Path | str | None, optional
        Ntuple directory where the h5 files are stored, by default None
    name: str | None, optional
        Name of the sample, for internal identification, by default None
    weights: list[float] | None, optional
        List of weights for this sample, by default None
    skip_checks: bool, optional
        Decide, if certain checks are skipped, by default False
    vds_dir: Path | str | None, optional
        Directory where virtual datasets will be stored if wildcard is used, by default None.
        If None, the virtual files will be created in the same directory as the input files.
    """

    pattern: Path | str | tuple[Path | str, ...]
    ntuple_dir: Path | str | None = None
    name: str | None = None
    weights: list[float] | None = None
    skip_checks: bool = False
    vds_dir: Path | str | None = None

    def __post_init__(self) -> None:
        if self.skip_checks:
            return
        if not self.pattern:
            raise ValueError("Sample pattern cannot be empty")
        if "*" in str(self.pattern) and not self.files:
            raise FileNotFoundError(f"No files matched pattern {self.pattern}")
        if missing := [file for file in self.files if not Path(file).exists()]:
            raise FileNotFoundError(f"The following files do not exist: {missing}")

    @property
    def path(self) -> tuple[Path, ...]:
        pattern_tuple = self.pattern if isinstance(self.pattern, list | tuple) else (self.pattern,)
        if self.ntuple_dir is not None:
            return tuple(Path(self.ntuple_dir, p) for p in pattern_tuple)
        return tuple(Path(p) for p in pattern_tuple)

    @property
    def files(self) -> list[str]:
        files = []
        for p in self.path:
            files += glob.glob(str(p)) if "*" in str(p) else [str(p)]
        return files

    @property
    def num_files(self) -> int:
        return len(self.files)

    @property
    def dsid(self) -> list[str]:
        return list({Path(fname).parent.name for fname in self.files})

    @property
    def sample_id(self) -> list[str]:
        return list({dsid.split(".")[2] for dsid in self.dsid})

    @property
    def tags(self) -> list[str]:
        return list({dsid.split(".")[3] for dsid in self.dsid})

    @property
    def ptag(self) -> list[str]:
        return list({tag for tags in self.tags for tag in tags.split("_") if "p" in tag})

    @property
    def rtag(self) -> list[str]:
        return list({tag for tags in self.tags for tag in tags.split("_") if "r" in tag})

    @property
    def dumper_tag(self) -> list[str]:
        hashes = [remove_suffix(dsid.split(".")[7], "_output") for dsid in self.dsid]
        return list(set(hashes))

    def virtual_file(self, **kwargs) -> list[Path | str]:
        out: list[Path | str] = []
        for filepath in self.path:
            if "*" not in str(filepath):
                out.append(filepath)
                continue

            if self.vds_dir is None:
                # current default behaviour (backwards compatible)
                out.append(create_virtual_file(filepath, **kwargs))
                continue

            p = Path(filepath)
            out_fname = Path(self.vds_dir) / f"{p.parent.name}_vds.h5"
            out.append(create_virtual_file(p, out_fname=out_fname, **kwargs))

        return out

    def __str__(self):
        return self.name

    def __lt__(self, other):
        return self.name < other.name

    def __eq__(self, other):
        return self.name == other.name



"""HDF5 package of the atlas-ftag-tools."""

from __future__ import annotations

from .h5add_col import h5_add_column
from .h5reader import H5Reader
from .h5utils import cast_dtype, get_dtype, join_structured_arrays, structured_from_dict
from .h5writer import H5Writer

__all__ = [
    "H5Reader",
    "H5Writer",
    "cast_dtype",
    "get_dtype",
    "h5_add_column",
    "join_structured_arrays",
    "structured_from_dict",
]



from __future__ import annotations

import argparse
from pathlib import Path
from typing import TYPE_CHECKING

import numpy as np
from scipy.optimize import minimize

from ftag.cli_utils import HelpFormatter
from ftag.cuts import Cuts
from ftag.hdf5 import H5Reader
from ftag.labels import LabelContainer
from ftag.utils import calculate_rejection, get_discriminant, logger, set_log_level

if TYPE_CHECKING:  # pragma: no cover
    from collections.abc import Sequence

    from ftag.labels import Label


def convert_dict(
    fraction_values: dict | np.ndarray,
    backgrounds: LabelContainer,
) -> np.ndarray | dict:
    """Convert the fraction values from dict to array or vice versa.

    Parameters
    ----------
    fraction_values : dict | np.ndarray
        Dict of array with the fraction values
    backgrounds : LabelContainer
        LabelContainer with the background flavours

    Returns
    -------
    np.ndarray | dict
        Array or dict with the fraction values

    Raises
    ------
    TypeError
        If the type of the input was wrong
    """
    if isinstance(fraction_values, dict):
        return np.array([fraction_values[iter_bkg.frac_str] for iter_bkg in backgrounds])

    if isinstance(fraction_values, np.ndarray):
        fraction_values = [
            float(frac_value / np.sum(fraction_values)) for frac_value in fraction_values
        ]

        return dict(
            zip([iter_bkg.frac_str for iter_bkg in backgrounds], fraction_values, strict=False)
        )

    raise TypeError(
        f"Only input of type `dict` or `np.ndarray` are accepted! You gave {type(fraction_values)}"
    )


def get_bkg_norm_dict(
    jets: np.ndarray,
    tagger: str,
    signal: Label,
    flavours: LabelContainer,
    working_point: float,
) -> dict:
    """Get the normalisation dict for the background flavours.

    Parameters
    ----------
    jets : np.ndarray
        Loaded jets
    tagger : str
        Name of the tagger
    signal : Label
        Label instance of the signal
    flavours : LabelContainer
        LabelContainer instance with all flavours used
    working_point : float
        Working point that is to be used

    Returns
    -------
    dict
        Background normalisation dict
    """
    # Init a dict for the bkg rejection norm values
    bkg_rej_norm = {}

    # Get the background classes
    backgrounds = flavours.backgrounds(signal)

    # Define a bool array if the jet is signal
    is_signal = signal.cuts(jets).idx

    # Loop over backgrounds
    for bkg in backgrounds:
        # Get the fraction value dict to maximize rejection for given class
        frac_dict_bkg = {
            iter_bkg.frac_str: 1 - (0.01 * len(backgrounds)) if iter_bkg == bkg else 0.01
            for iter_bkg in backgrounds
        }

        # Calculate the disc value using the new fraction dict
        disc = get_discriminant(
            jets=jets,
            tagger=tagger,
            signal=signal,
            flavours=flavours,
            fraction_values=frac_dict_bkg,
        )

        # Calculate the discriminant
        bkg_rej_norm[bkg.name] = calculate_rejection(
            sig_disc=disc[is_signal],
            bkg_disc=disc[bkg.cuts(jets).idx],
            target_eff=working_point,
        )

    return bkg_rej_norm


def calculate_rejection_sum(
    fraction_dict: dict | np.ndarray,
    jets: np.ndarray,
    tagger: str,
    signal: Label,
    flavours: LabelContainer,
    working_point: float,
    bkg_norm_dict: dict,
    rejection_weights: dict,
) -> float:
    """Calculate the sum of the normalised rejections.

    Parameters
    ----------
    fraction_dict : dict | np.ndarray
        Dict/Array with the fraction values
    jets : np.ndarray
        Loaded jets
    tagger : str
        Name of the tagger
    signal : Label
        Label instance of the signal
    flavours : LabelContainer
        LabelContainer with all flavours
    working_point : float
        Working point that is used
    bkg_norm_dict : dict
        Backgroud normalisation dict
    rejection_weights : dict
        Weights for the rejections

    Returns
    -------
    float
        Sum of the normalised rejections times -1 (for minimize)
    """
    # Get the background classes
    backgrounds = flavours.backgrounds(signal)

    # Define a bool array if the jet is signal
    is_signal = signal.cuts(jets).idx

    # Check that the fraction dict is a dict
    if isinstance(fraction_dict, np.ndarray):
        fraction_dict = convert_dict(
            fraction_values=fraction_dict,
            backgrounds=backgrounds,
        )

    # Calculate discriminant
    disc = get_discriminant(
        jets=jets,
        tagger=tagger,
        signal=signal,
        flavours=flavours,
        fraction_values=fraction_dict,
    )

    # Init a dict to which the bkg rejs are added
    sum_bkg_rej = 0

    # Loop over the backgrounds and calculate the rejections
    for iter_bkg in backgrounds:
        sum_bkg_rej += (
            calculate_rejection(
                sig_disc=disc[is_signal],
                bkg_disc=disc[iter_bkg.cuts(jets).idx],
                target_eff=working_point,
            )
            / bkg_norm_dict[iter_bkg.name]
        ) * rejection_weights[iter_bkg.name]

    # Return the negative sum to enable minimizer
    return -1 * sum_bkg_rej


def calculate_best_fraction_values(
    jets: np.ndarray,
    tagger: str,
    signal: Label,
    flavours: LabelContainer,
    working_point: float,
    rejection_weights: dict | None = None,
    optimizer_method: str = "Powell",
) -> dict:
    """Calculate the best fraction values for a given tagger and working point.

    Parameters
    ----------
    jets : np.ndarray
        Loaded jets
    tagger : str
        Name of the tagger
    signal : Label
        Label instance of the signal
    flavours : LabelContainer
        LabelContainer with all flavours
    working_point : float
        Working point that is used
    rejection_weights : dict | None, optional
        Rejection weights for the background classes, by default None
    optimizer_method : str, optional
        Optimizer method for the minimization, by default "Powell"

    Returns
    -------
    dict
        Dict with the best fraction values
    """
    logger.debug("Calculating best fraction values.")
    logger.debug(f"tagger: {tagger}")
    logger.debug(f"signal: {signal}")
    logger.debug(f"flavours: {flavours}")
    logger.debug(f"working_point: {working_point}")
    logger.debug(f"rejection_weights: {rejection_weights}")
    logger.debug(f"optimizer_method: {optimizer_method}")

    # Ensure Label instance
    if isinstance(signal, str):
        signal = flavours[signal]

    # Get the background classes
    backgrounds = flavours.backgrounds(signal)

    # Define a default fraction dict
    def_frac_dict = {iter_bkg.frac_str: 1 / len(backgrounds) for iter_bkg in backgrounds}

    # Define rejection weights if not set
    if rejection_weights is None:
        rejection_weights = {iter_bkg.name: 1 for iter_bkg in backgrounds}

    # Get the normalisation for all bkg rejections
    bkg_norm_dict = get_bkg_norm_dict(
        jets=jets,
        tagger=tagger,
        signal=signal,
        flavours=flavours,
        working_point=working_point,
    )

    # Get the best fraction values combination
    result = minimize(
        fun=calculate_rejection_sum,
        x0=convert_dict(fraction_values=def_frac_dict, backgrounds=backgrounds),
        method=optimizer_method,
        bounds=[(0, 1)] * len(backgrounds),
        args=(jets, tagger, signal, flavours, working_point, bkg_norm_dict, rejection_weights),
    )

    # Get the final fraction dict
    final_frac_dict = convert_dict(fraction_values=result.x, backgrounds=backgrounds)

    logger.info(f"Minimization Success: {result.success}")
    logger.info("The following best fraction values were found:")
    for frac_str, frac_value in final_frac_dict.items():
        logger.info(f"{frac_str}: {round(frac_value, ndigits=3)}")

    return final_frac_dict


def parse_args(args: Sequence[str] | None) -> argparse.Namespace:
    """Parse the input arguments into a Namespace.

    Parameters
    ----------
    args : Sequence[str] | None
        Sequence of string inputs to the script

    Returns
    -------
    argparse.Namespace
        Namespace with the parsed arguments
    """
    # Create the parser
    parser = argparse.ArgumentParser(
        description=__doc__,
        formatter_class=HelpFormatter,
    )
    parser.add_argument(
        "-i",
        "--input_file",
        required=True,
        type=Path,
        help="Path to the H5 file that will be used. Wildcard for multiple H5 files is supported",
    )
    parser.add_argument(
        "-t",
        "--tagger",
        required=True,
        type=str,
        help="Name of the tagger in the files.",
    )
    parser.add_argument(
        "-s",
        "--signal",
        default="bjets",
        type=str,
        help="Name of the signal class.",
    )
    parser.add_argument(
        "-w",
        "--working_point",
        required=True,
        type=float,
        help="Working point for which to optimize the fraction values. "
        "Values are given in sub-one values (e.g. 0.70 for 70%% WP).",
    )
    parser.add_argument(
        "-o",
        "--optimizer_method",
        default="Powell",
        type=str,
        help="Optimizer method for the minimization.",
    )
    parser.add_argument(
        "-n",
        "--num_jets",
        default=None,
        type=int,
        help="Number of jets to load from H5. By default None (all jets).",
    )
    parser.add_argument(
        "-c",
        "--cuts",
        default=[
            "pt_btagJes > 20e3",
            "pt_btagJes < 250e3",
            "absEta_btagJes < 2.5",
        ],
        type=list,
        help="Cuts that are to be applied as list. Default is the ttbar selection we use. ",
    )
    parser.add_argument(
        "--batch_size",
        default=100_000,
        type=int,
        help="Batch size used when loading the jets from H5.",
    )
    parser.add_argument(
        "--jets_name",
        default="jets",
        type=str,
        help="Name of the jet collection in the H5 file.",
    )
    parser.add_argument(
        "--flavour_file",
        default=None,
        type=str,
        help="File with custom flavour definition that differs from the flavours.yaml",
    )
    parser.add_argument(
        "--flavour_class",
        default="single-btag",
        type=str,
        help="Name of the category of the flavours that will be used.",
    )
    parser.add_argument(
        "-v",
        "--verbose",
        default="INFO",
        choices=["DEBUG", "INFO", "WARNING", "ERROR", "CRITICAL"],
        help="Verbosity level",
    )

    # Final parse of all arguments
    return parser.parse_args(args)


def main(args: Sequence[str] | None = None) -> None:
    """Main function to run the fraction value optimization.

    Parameters
    ----------
    args : Sequence[str] | None
        Input arguments, by default None
    """
    # Parse the command line arguments
    parsed_args = parse_args(args=args)

    # Set the log level
    set_log_level(tools_logger=logger, log_level=parsed_args.verbose)

    # Get the flavour class we will use
    flavours = LabelContainer.from_yaml(
        yaml_path=parsed_args.flavour_file,
        include_categories=[parsed_args.flavour_class],
    )

    # Get the cuts that are to be applied
    cuts = Cuts.from_list(parsed_args.cuts)

    # Get the probability names which need to be loaded from H5
    vars_to_load = [f"{parsed_args.tagger}_{flav.px}" for flav in flavours]

    # Add the variables needed for the flavours to the list
    vars_to_load += flavours.cut_variables()

    # Ensure that each variable only is once in the list
    vars_to_load = list(set(vars_to_load))

    # Debug statement before loading the jets
    logger.debug(f"Flavours: {flavours}")
    logger.debug(f"Flavour File: {parsed_args.flavour_file}")
    logger.debug(f"Variables: {vars_to_load}")
    logger.debug(f"Cuts: {vars_to_load}")

    # Load the actual jets from the file
    jets = H5Reader(
        fname=parsed_args.input_file,
        batch_size=parsed_args.batch_size,
        jets_name=parsed_args.jets_name,
        shuffle=False,
    ).load(
        variables={parsed_args.jets_name: vars_to_load},
        num_jets=parsed_args.num_jets,
        cuts=cuts,
    )[parsed_args.jets_name]

    # Call the actual fraction value optimization function
    calculate_best_fraction_values(
        jets=jets,
        tagger=parsed_args.tagger,
        signal=parsed_args.signal,
        flavours=flavours,
        working_point=parsed_args.working_point,
        optimizer_method=parsed_args.optimizer_method,
    )


if __name__ == "__main__":  # pragma: no cover
    main()



from __future__ import annotations

from collections.abc import Callable
from dataclasses import dataclass, field

import numpy as np

__all__ = ["Transform"]

Batch = dict[str, np.ndarray]


@dataclass
class Transform:
    """Apply variable name remapping, integer remapping, and float transformations.

    The Transform class provides a unified mechanism to perform:
    - variable renaming (variable_map)
    - integer value remapping (ints_map)
    - float transformations (floats_map)

    Each transformation is applied to a *batch* consisting of a dictionary of
    structured numpy arrays.

    Attributes
    ----------
    variable_map : dict[str, dict[str, str]]
        A nested mapping where variable_map[group][old] = new specifies how
        variable names should be renamed inside a given group. If None, no
        variable renaming is applied.

    ints_map : dict[str, dict[str, dict[int, int]]]
        A nested mapping where ints_map[group][variable][old] = new specifies
        how integer values should be remapped. If None, no integer remapping
        is applied.

    floats_map : dict[str, dict[str, str | Callable]]
        A nested mapping where floats_map[group][variable] = func specifies
        a float transformation function. func may either be:
        - a callable
        - a string giving the name of a numpy function (e.g. "log")

        Strings are resolved to numpy.<func> automatically.

    variable_map_inv : dict[str, dict[str, str]]
        Automatically generated inverse of variable_map used for reverse
        variable lookup in :meth:`map_variable_names`.
    """

    variable_map: dict[str, dict[str, str]] = field(default_factory=dict)
    ints_map: dict[str, dict[str, dict[int, int]]] = field(default_factory=dict)
    floats_map: dict[str, dict[str, str | Callable]] = field(default_factory=dict)

    variable_map_inv: dict[str, dict[str, str]] = field(init=False)

    def __post_init__(self) -> None:
        """Initialize internal maps and convert float transformation strings.

        This method ensures that variable_map, ints_map, and
        floats_map are always dictionaries (never None), constructs the
        inverse variable map, and converts any string-based float
        transformations into their numpy equivalents.
        """
        # Normalize maps
        self.variable_map = self.variable_map or {}
        self.ints_map = self.ints_map or {}
        self.floats_map = self.floats_map or {}

        # Inverse variable mapping
        self.variable_map_inv = {
            group: {new: old for old, new in mapping.items()}
            for group, mapping in self.variable_map.items()
        }

        # Convert float-string mappings to numpy functions
        for group, map_dict in self.floats_map.items():
            for variable, func in map_dict.items():
                if isinstance(func, str):
                    self.floats_map[group][variable] = getattr(np, func)

    def __call__(self, batch: Batch) -> Batch:
        """Apply integer remapping, float transformations, and variable renaming.

        Parameters
        ----------
        batch : Batch
            A mapping from group name to structured numpy arrays.

        Returns
        -------
        Batch
            The transformed batch.
        """
        batch = self.map_ints(batch)
        batch = self.map_floats(batch)
        return self.map_variables(batch)

    def map_variables(self, batch: Batch) -> Batch:
        """Rename variables in each group according to variable_map.

        Parameters
        ----------
        batch : Batch
            Dictionary mapping group names to structured numpy arrays.

        Returns
        -------
        Batch
            The batch with variables renamed where applicable.
        """
        assert self.variable_map is not None
        for group in self.variable_map:
            if group in batch:
                batch[group] = batch[group].astype(self.map_dtype(group, batch[group].dtype))
        return batch

    def map_ints(self, batch: Batch) -> Batch:
        """Remap integer values for specified variables inside each group.

        Parameters
        ----------
        batch : Batch
            Dictionary mapping group names to structured numpy arrays.

        Returns
        -------
        Batch
            The batch with integer values remapped.
        """
        assert self.ints_map is not None
        for group, map_dict in self.ints_map.items():
            if group not in batch:
                continue
            for variable, int_map in map_dict.items():
                if variable not in (batch[group].dtype.names or ()):
                    continue
                data = batch[group][variable]
                for old, new in int_map.items():
                    data[data == old] = new
        return batch

    def map_floats(self, batch: Batch) -> Batch:
        """Apply float transformations to selected variables.

        Parameters
        ----------
        batch : Batch
            Dictionary mapping group names to structured numpy arrays.

        Returns
        -------
        Batch
            The batch with float transformations applied.
        """
        assert self.floats_map is not None
        for group, map_dict in self.floats_map.items():
            if group not in batch:
                continue
            for variable, func in map_dict.items():
                assert callable(func)
                batch[group][variable] = func(batch[group][variable])
        return batch

    def map_dtype(self, name: str, dtype: np.dtype) -> np.dtype:
        """Compute a new dtype with renamed fields according to variable_map.

        Parameters
        ----------
        name : str
            Group name associated with the dtype.
        dtype : np.dtype
            Structured dtype whose field names may be modified.

        Returns
        -------
        np.dtype
            A dtype with renamed fields where required.

        Raises
        ------
        ValueError
            When the variables already exist in the dataset.
        """
        assert self.variable_map is not None
        if not (map_dict := self.variable_map.get(name.lstrip("/"))):
            return dtype

        names = list(dtype.names or ())
        for old, new in map_dict.items():
            if old in names and new in names:
                raise ValueError(f"Variables {old, new} already exist in {name}.")
        return np.dtype([(map_dict.get(field, field), dtype[field]) for field in names])

    def map_variable_names(
        self,
        name: str,
        variables: list[str],
        inverse: bool = False,
    ) -> list[str]:
        """Map a list of variable names using variable_map or variable_map_inv.

        Parameters
        ----------
        name : str
            Group name used to select the appropriate name-mapping dictionary.
        variables : list[str]
            List of variable names to be mapped.
        inverse : bool, optional
            If False (default), apply variable_map.
            If True, apply the inverse mapping variable_map_inv.

        Returns
        -------
        list[str]
            A new list of mapped variable names.
        """
        key = name.lstrip("/")

        # Define the map dict
        map_dict = self.variable_map_inv.get(key) if inverse else self.variable_map.get(key)

        # Return variables if map_dict is None
        if not map_dict:
            return variables

        return [map_dict.get(var, var) for var in variables]



"""Tools for metrics module."""

from __future__ import annotations

from typing import TYPE_CHECKING

import numpy as np
from scipy.ndimage import gaussian_filter1d

from ftag.utils import logger

if TYPE_CHECKING:  # pragma: no cover
    from ftag.labels import Label, LabelContainer


def save_divide(
    numerator: np.ndarray | float,
    denominator: np.ndarray | float,
    default: float = 1.0,
) -> np.ndarray | float:
    """Save divide for denominator equal to 0.

    Division using numpy divide function returning default value in cases where
    denominator is 0.

    Parameters
    ----------
    numerator: np.ndarray | float
        Numerator in the ratio calculation.
    denominator: np.ndarray | float
        Denominator in the ratio calculation.
    default: float
        Default value which is returned if denominator is 0.

    Returns
    -------
    np.ndarray | float
        Result of the division
    """
    logger.debug("Calculating save division.")
    logger.debug("numerator: %s", numerator)
    logger.debug("denominator: %s", denominator)
    logger.debug("default: %s", default)

    if isinstance(numerator, int | float | np.number) and isinstance(
        denominator, int | float | np.number
    ):
        output_shape = 1
    else:
        try:
            output_shape = denominator.shape
        except AttributeError:
            output_shape = numerator.shape

    ratio = np.divide(
        numerator,
        denominator,
        out=np.ones(
            output_shape,
            dtype=float,
        )
        * default,
        where=(denominator != 0),
    )
    if output_shape == 1:
        return float(ratio.item())
    return ratio


def weighted_percentile(
    arr: np.ndarray,
    percentile: np.ndarray,
    weights: np.ndarray = None,
) -> np.ndarray:
    """Calculate weighted percentile.

    Implementation according to https://stackoverflow.com/a/29677616/11509698
    (https://en.wikipedia.org/wiki/Percentile#The_weighted_percentile_method)

    Parameters
    ----------
    arr : np.ndarray
        Data array
    percentile : np.ndarray
        Percentile array
    weights : np.ndarray
        Weights array, by default None

    Returns
    -------
    np.ndarray
        Weighted percentile array
    """
    logger.debug("Calculating weighted percentile.")
    logger.debug("arr: %s", arr)
    logger.debug("percentile: %s", percentile)
    logger.debug("weights: %s", weights)

    # Set weights to one if no weights are given
    if weights is None:
        weights = np.ones_like(arr)

    # Set dtype to float64 if the weights are too large
    dtype = np.float64 if np.sum(weights) > 1000000 else np.float32

    # Get an array sorting and sort the array and the weights
    ix = np.argsort(arr)
    arr = arr[ix]
    weights = weights[ix]

    # Return the cumulative sum
    cdf = np.cumsum(weights, dtype=dtype) - 0.5 * weights
    cdf -= cdf[0]
    cdf /= cdf[-1]

    # Return the linear interpolation
    return np.interp(percentile, cdf, arr)


def calculate_efficiency(
    sig_disc: np.ndarray,
    bkg_disc: np.ndarray,
    target_eff: float | list | np.ndarray | None = None,
    cut_value: float | list | np.ndarray | None = None,
    return_cuts: bool = False,
    sig_weights: np.ndarray = None,
    bkg_weights: np.ndarray = None,
):
    """Calculate efficiency.

    Parameters
    ----------
    sig_disc : np.ndarray
        Signal discriminant
    bkg_disc : np.ndarray
        Background discriminant
    target_eff : float | list | np.ndarray | None, optional
        Working point which is used for discriminant calculation. By default None
    cut_value : float | list | np.ndarray | None, optional
        Instead of the working points being calculated on the sample, you can give
        provide the cut values directly. By default None
    return_cuts : bool
        Specifies if the cut values corresponding to the provided WPs are returned.
        If target_eff is a float, only one cut value will be returned. If target_eff
        is an array, target_eff is an array as well.
    sig_weights : np.ndarray
        Weights for signal events
    bkg_weights : np.ndarray
        Weights for background events

    Returns
    -------
    eff : float or np.ndarray
        Efficiency.
        Return float if target_eff is a float, else np.ndarray
    cut_vals : float or np.ndarray
        Cutvalue if return_cuts is True.
        Return float if target_eff is a float, else np.ndarray

    Raises
    ------
    ValueError
        If target_eff and cut_value are both not None or both are None
    """
    logger.debug("Calculating efficiency.")
    logger.debug("sig_disc: %s", sig_disc)
    logger.debug("bkg_disc: %s", bkg_disc)
    logger.debug("target_eff: %s", target_eff)
    logger.debug("return_cuts: %s", return_cuts)
    logger.debug("sig_weights: %s", sig_weights)
    logger.debug("bkg_weights: %s", bkg_weights)

    if (target_eff is None) == (cut_value is None):
        raise ValueError("Either target_eff or cut_value must/can be defined!")

    # float | np.ndarray for both target_eff and the returned values
    return_float = False
    if isinstance(target_eff, float):
        return_float = True

    # Calculate the cut_vals based on the target_eff
    if target_eff is not None:
        # Flatten the target efficiencies
        flat_target_eff: np.ndarray = np.asarray(target_eff, dtype=float).ravel()

        # Get the cut_vals for the given target efficiency
        cut_vals = weighted_percentile(
            arr=sig_disc,
            percentile=1.0 - flat_target_eff,
            weights=sig_weights,
        )

        # Sort the cut_vals to get the correct order
        sorted_args = np.argsort(1 - flat_target_eff)

    # If the cut values are provided, use them
    else:
        cut_vals = np.asarray([cut_value]).flatten()

        # Sort the cut_vals to get the correct order
        sorted_args = np.argsort(cut_vals)

    # Get the histogram for the backgrounds
    hist, _ = np.histogram(bkg_disc, (-np.inf, *cut_vals[sorted_args], np.inf), weights=bkg_weights)

    # Calculate the efficiencies for the calculated cut values
    eff = hist[::-1].cumsum()[-2::-1] / hist.sum()
    eff = eff[sorted_args]

    # Ensure that a float is returned if float was given
    if return_float:
        eff = eff[0]
        cut_vals = cut_vals[0]

    # Also return the cuts if wanted
    if return_cuts:
        return eff, cut_vals

    return eff


def calculate_rejection(
    sig_disc: np.ndarray,
    bkg_disc: np.ndarray,
    target_eff: float | list | None = None,
    cut_value: float | list | None = None,
    return_cuts: bool = False,
    sig_weights: np.ndarray = None,
    bkg_weights: np.ndarray = None,
    smooth: bool = False,
):
    """Calculate rejection.

    Parameters
    ----------
    sig_disc : np.ndarray
        Signal discriminant
    bkg_disc : np.ndarray
        Background discriminant
    target_eff : float | list | None, optional
        Working point which is used for discriminant calculation. By default None
    cut_value : float | list | None, optional
        Instead of the working points being calculated on the sample, you can give
        provide the cut values directly. By default None
    return_cuts : bool, optional
        Specifies if the cut values corresponding to the provided WPs are returned.
        If target_eff is a float, only one cut value will be returned. If target_eff
        is an array, target_eff is an array as well.
    sig_weights : np.ndarray, optional
        Weights for signal events, by default None
    bkg_weights : np.ndarray, optional
        Weights for background events, by default None
    smooth: bool, optional
        Decide, if the output is smoothed using a 1D gaussian filter. By default False

    Returns
    -------
    rej : float or np.ndarray
        Rejection.
        If target_eff is a float, a float is returned if it's a list a np.ndarray
    cut_value : float or np.ndarray
        Cutvalue if return_cuts is True.
        If target_eff is a float, a float is returned if it's a list a np.ndarray

    Raises
    ------
    ValueError
        If target_eff and cut_value are both not None or both are None
    """
    logger.debug("Calculating rejection.")
    logger.debug("sig_disc: %s", sig_disc)
    logger.debug("bkg_disc: %s", bkg_disc)
    logger.debug("target_eff: %s", target_eff)
    logger.debug("return_cuts: %s", return_cuts)
    logger.debug("sig_weights: %s", sig_weights)
    logger.debug("bkg_weights: %s", bkg_weights)
    logger.debug("smooth: %s", smooth)

    if (target_eff is None) == (cut_value is None):
        raise ValueError("Either target_eff or cut_value must/can be defined!")

    # Calculate the efficiency
    eff = calculate_efficiency(
        sig_disc=sig_disc,
        bkg_disc=bkg_disc,
        target_eff=target_eff,
        cut_value=cut_value,
        return_cuts=return_cuts,
        sig_weights=sig_weights,
        bkg_weights=bkg_weights,
    )

    # Invert the efficiency to get a rejection
    rej = save_divide(1, eff[0] if return_cuts else eff, np.inf)

    # Smooth out the rejection if wanted
    if smooth:
        rej = gaussian_filter1d(rej, sigma=1, radius=2, mode="nearest")

    # Return also the cut values if wanted
    if return_cuts:
        return rej, eff[1]

    return rej


def calculate_efficiency_error(
    arr: np.ndarray,
    n_counts: int,
    suppress_zero_divison_error: bool = False,
    norm: bool = False,
) -> np.ndarray:
    """Calculate statistical efficiency uncertainty.

    Parameters
    ----------
    arr : np.ndarray
        Efficiency values
    n_counts : int
        Number of used statistics to calculate efficiency
    suppress_zero_divison_error : bool
        Not raising Error for zero division
    norm : bool, optional
        If True, normed (relative) error is being calculated, by default False

    Returns
    -------
    np.ndarray
        Efficiency uncertainties

    Raises
    ------
    ValueError
        If n_counts <=0

    Notes
    -----
    This method uses binomial errors as described in section 2.2 of
    https://inspirehep.net/files/57287ac8e45a976ab423f3dd456af694
    """
    logger.debug("Calculating efficiency error.")
    logger.debug("arr: %s", arr)
    logger.debug("n_counts: %i", n_counts)
    logger.debug("suppress_zero_divison_error: %s", suppress_zero_divison_error)
    logger.debug("norm: %s", norm)
    if np.any(n_counts <= 0) and not suppress_zero_divison_error:
        raise ValueError(f"You passed as argument `N` {n_counts} but it has to be larger 0.")
    if norm:
        return np.sqrt(arr * (1 - arr) / n_counts) / arr
    return np.sqrt(arr * (1 - arr) / n_counts)


def calculate_rejection_error(
    arr: np.ndarray,
    n_counts: int,
    norm: bool = False,
) -> np.ndarray:
    """Calculate the rejection uncertainties.

    Parameters
    ----------
    arr : np.ndarray
        Rejection values
    n_counts : int
        Number of used statistics to calculate rejection
    norm : bool, optional
        If True, normed (relative) error is being calculated, by default False

    Returns
    -------
    np.ndarray
        Rejection uncertainties

    Raises
    ------
    ValueError
        If n_counts <=0 or if the rejection value is 0

    Notes
    -----
    Special case of `eff_err()`
    """
    logger.debug("Calculating rejection error.")
    logger.debug("arr: %s", arr)
    logger.debug("n_counts: %i", n_counts)
    logger.debug("norm: %s", norm)
    if np.any(n_counts <= 0):
        raise ValueError(f"You passed as argument `n_counts` {n_counts} but it has to be larger 0.")
    if np.any(arr == 0):
        raise ValueError("One rejection value is 0, cannot calculate error.")
    if norm:
        return np.power(arr, 2) * calculate_efficiency_error(1 / arr, n_counts) / arr
    return np.power(arr, 2) * calculate_efficiency_error(1 / arr, n_counts)


def get_discriminant(
    jets: np.ndarray,
    tagger: str,
    signal: Label,
    flavours: LabelContainer,
    fraction_values: dict[str, float],
    epsilon: float = 1e-10,
) -> np.ndarray:
    """Calculate the tagging discriminant for a given tagger.

    Calculated as the logarithm of the ratio of a specified signal probability
    to a weighted sum ofbackground probabilities.

    Parameters
    ----------
    jets : np.ndarray
        Structured array of jets containing tagger outputs
    tagger : str
        Name of the tagger
    signal : Label
        Signal flavour (bjets/cjets or hbb/hcc)
    flavours: LabelContainer
        Labels that are used in the discriminant calculation
    fraction_values : dict[str, float]
        Dict with the fraction values for the background classes for the given tagger
    epsilon : float, optional
        Small number to avoid division by zero, by default 1e-10

    Returns
    -------
    np.ndarray
        Array of discriminant values.

    Raises
    ------
    ValueError
        If the signal flavour is not recognised.
    """
    # Init the denominator
    denominator = 0.0

    # Loop over background flavours
    for flav in flavours:
        # Skip signal flavour for denominator
        if flav == signal:
            continue

        # Get the probability name of the tagger/flavour combo + fraction value
        prob_name = f"{tagger}_{flav.px}"
        fraction_value = fraction_values[flav.frac_str]

        # If fraction_value for the given flavour is zero, skip it
        if fraction_value == 0:
            continue

        # Check that the probability value for the flavour is available
        if fraction_value > 0 and prob_name not in jets.dtype.names:
            raise ValueError(
                f"Nonzero fraction value for {flav.name}, but '{prob_name}' "
                "not found in input array."
            )

        # Update denominator
        denominator += jets[prob_name] * fraction_value if prob_name in jets.dtype.names else 0

    # Calculate numerator
    signal_field = f"{tagger}_{signal.px}"

    # Check that the probability of the signal is available
    if signal_field not in jets.dtype.names:
        raise ValueError(
            f"No signal probability value(s) found for tagger {tagger}. "
            f"Missing variable: {signal_field}"
        )

    return np.log((jets[signal_field] + epsilon) / (denominator + epsilon))



from __future__ import annotations

import argparse
import glob
import re
import sys
from pathlib import Path
from typing import Any

import h5py
import numpy as np


def parse_args(args: Any | None) -> argparse.Namespace:
    """Parse command line arguments.

    Parameters
    ----------
    args : Any | None
        Command line arguments

    Returns
    -------
    argparse.Namespace
        Namespace with the parsed command line arguments.
    """
    parser = argparse.ArgumentParser(
        description="Create a lightweight HDF5 wrapper (virtual datasets + "
        "summed cutBookkeeper counts) around a set of .h5 files"
    )
    parser.add_argument(
        "pattern",
        type=Path,
        help="quotes-enclosed glob pattern of files to merge, or a regex if --use_regex is given",
    )
    parser.add_argument("output", type=Path, help="path to output virtual file")
    parser.add_argument(
        "--use_regex",
        action="store_true",
        help="treat PATTERN as a regular expression instead of a glob",
    )
    parser.add_argument(
        "--regex_path",
        type=str,
        required="--use_regex" in (args or sys.argv),
        default=None,
        help="directory whose entries the regex is applied to "
        "(defaults to the current working directory)",
    )
    return parser.parse_args(args)


def get_virtual_layout(fnames: list[str], group: str) -> h5py.VirtualLayout:
    """Concatenate group from multiple files into a single VirtualDataset.

    Parameters
    ----------
    fnames : list[str]
        List with the file names
    group : str
        Name of the group that is concatenated

    Returns
    -------
    h5py.VirtualLayout
        Virtual layout of the new virtual dataset
    """
    sources = []
    total = 0

    # Loop over the input files
    for fname in fnames:
        with h5py.File(fname, "r") as f:
            # Get the file and append its length
            vsrc = h5py.VirtualSource(f[group])
            total += vsrc.shape[0]
            sources.append(vsrc)

    # Define the layout of the output vds
    with h5py.File(fnames[0], "r") as f:
        dtype = f[group].dtype
        shape = f[group].shape

    # Update the shape finalize the output layout
    shape = (total, *shape[1:])
    layout = h5py.VirtualLayout(shape=shape, dtype=dtype)

    # Fill the vds
    idx = 0
    for vsrc in sources:
        length = vsrc.shape[0]
        layout[idx : idx + length] = vsrc
        idx += length

    return layout


def glob_re(pattern: str | None, regex_path: str | None) -> list[str] | None:
    """Return list of filenames that match a regex pattern inside `regex_path`.

    Parameters
    ----------
    pattern : str | None
        Regex pattern used to match filenames.
    regex_path : str | None
        Directory to search for files.

    Returns
    -------
    list[str] | None
        List of basenames that match the pattern, or None if inputs are missing.
    """
    if pattern is None or regex_path is None:
        return None

    path = Path(regex_path)
    regex = re.compile(pattern)

    # Return only basenames, same as os.listdir
    return [p.name for p in path.iterdir() if regex.match(p.name)]


def regex_files_from_dir(
    reg_matched_fnames: list[str] | None,
    regex_path: str | None,
) -> list[str] | None:
    """Turn a list of basenames into full paths; dive into sub-dirs if needed.

    Parameters
    ----------
    reg_matched_fnames : list[str] | None
        List of the regex matched file names
    regex_path : str | None
        Regex path for the input files

    Returns
    -------
    list[str] | None
        List of file paths (as strings) that matched the regex and any subsequent
        globbing inside matched directories.
    """
    if reg_matched_fnames is None or regex_path is None:
        return None

    parent_dir = regex_path or str(Path.cwd())
    full_paths = [Path(parent_dir) / fname for fname in reg_matched_fnames]
    paths_to_glob = [str(fp / "*.h5") if fp.is_dir() else str(fp) for fp in full_paths]
    nested_fnames = [glob.glob(p) for p in paths_to_glob]
    return sum(nested_fnames, [])


def sum_counts_once(counts: np.ndarray) -> np.ndarray:
    """Reduce the arrays in the counts dataset for one file to a scalar via summation.

    Parameters
    ----------
    counts : np.ndarray
        Array from the h5py dataset (counts) from the cutBookkeeper groups

    Returns
    -------
    np.ndarray
        Array with the summed variables for the file
    """
    dtype = counts.dtype
    summed = np.zeros((), dtype=dtype)
    for field in dtype.names:
        summed[field] = counts[field].sum()
    return summed


def check_subgroups(fnames: list[str], group_name: str = "cutBookkeeper") -> list[str]:
    """Check which subgroups are available for the bookkeeper.

    Find the intersection of sub-group names that have a 'counts' dataset
    in every input file. (Using the intersection makes the script robust
    even if a few files are missing a variation.)

    Parameters
    ----------
    fnames : list[str]
        List of the input files
    group_name : str, optional
        Group name in the h5 files of the bookkeeper, by default "cutBookkeeper"

    Returns
    -------
    list[str]
        Returns the files with common sub-groups

    Raises
    ------
    KeyError
        When a file does not have a bookkeeper
    ValueError
        When no common bookkeeper sub-groups were found
    """
    common: set[str] | None = None
    for fname in fnames:
        with h5py.File(fname, "r") as f:
            if group_name not in f:
                raise KeyError(f"{fname} has no '{group_name}' group")
            these = {
                name
                for name, item in f[group_name].items()
                if isinstance(item, h5py.Group) and "counts" in item
            }
            common = these if common is None else common & these
    if not common:
        raise ValueError("No common cutBookkeeper sub-groups with 'counts' found")
    return sorted(common)


def aggregate_cutbookkeeper(
    fnames: list[str],
    group_name: str = "cutBookkeeper",
) -> dict[str, np.ndarray] | None:
    """Aggregate the cutBookkeeper in the input files.

    For every input file:
    For every sub-group (nominal, sysUp, sysDown, …):
    1. Sum the 4-entry record array inside each file into 1 record
    1. Add those records from all files together into grand total
    Returns a dict  {subgroup_name: scalar-record-array}

    Parameters
    ----------
    fnames : list[str]
        List of the input files
    group_name: str, optional
        Group name of the cutBookkeeper. By default "cutBookkeeper"

    Returns
    -------
    dict[str, np.ndarray] | None
        Dict with the accumulated cutBookkeeper groups. If the cut bookkeeper
        is not in the files, return None.
    """
    if any(group_name not in h5py.File(f, "r") for f in fnames):
        return None

    subgroups = check_subgroups(fnames, group_name=group_name)

    # initialise an accumulator per subgroup (dtype taken from 1st file)
    accum: dict[str, np.ndarray] = {}
    with h5py.File(fnames[0], "r") as f0:
        for sg in subgroups:
            dtype = f0[f"{group_name}/{sg}/counts"].dtype
            accum[sg] = np.zeros((), dtype=dtype)

    # add each files contribution field-wise
    for fname in fnames:
        with h5py.File(fname, "r") as f:
            for sg in subgroups:
                per_file = sum_counts_once(f[f"{group_name}/{sg}/counts"][()])
                for fld in accum[sg].dtype.names:
                    accum[sg][fld] += per_file[fld]

    return accum


def create_virtual_file(
    pattern: Path | str,
    out_fname: Path | str | None = None,
    use_regex: bool = False,
    regex_path: str | None = None,
    overwrite: bool = False,
    bookkeeper_name: str = "cutBookkeeper",
) -> Path:
    """Create the virtual dataset file for the given inputs.

    Parameters
    ----------
    pattern : Path | str
        Pattern of the input files used. Wildcard is supported
    out_fname : Path | str | None, optional
        Output path to which the virtual dataset file is written. By default None
    use_regex : bool, optional
        If you want to use regex instead of glob, by default False
    regex_path : str | None, optional
        Regex logic used to define the input files, by default None
    overwrite : bool, optional
        Decide, if an existing output file is overwritten, by default False
    bookkeeper_name : str, optional
        Name of the cut bookkeeper in the h5 files.

    Returns
    -------
    Path
        Path object of the path to which the output file is written

    Raises
    ------
    FileNotFoundError
        If not input files were found for the given pattern
    ValueError
        If no output file is given and the input comes from multiple directories
    """
    # Get list of filenames
    pattern_str = str(pattern)

    # Use regex to find input files else use glob
    if use_regex is True:
        matched = glob_re(pattern_str, regex_path)
        fnames = regex_files_from_dir(matched, regex_path)
    else:
        fnames = glob.glob(pattern_str)

    # Throw error if no input files were found
    if not fnames:
        raise FileNotFoundError(f"No files matched pattern {pattern!r}")

    # Infer output path if not given
    if out_fname is None:
        if len({Path(f).parent for f in fnames}) != 1:
            raise ValueError("Give --output when files reside in multiple dirs")
        out_fname = Path(fnames[0]).parent / "vds" / "vds.h5"
    else:
        out_fname = Path(out_fname)

    # If overwrite is not active and a file exists, stop here
    if not overwrite and out_fname.is_file():
        return out_fname

    # Identify common groups across all files
    common_groups: set[str] = set()
    for fname in fnames:
        with h5py.File(fname, "r") as f:
            groups = set(f.keys())
            common_groups = groups if not common_groups else common_groups & groups

    # Ditch the bookkeeper. We will process it separately
    common_groups.discard("cutBookkeeper")

    # Check that the directory of the output file exists
    out_fname.parent.mkdir(parents=True, exist_ok=True)

    # Build the output file
    with h5py.File(out_fname, "w") as fout:
        # Build "standard" groups
        for gname in sorted(common_groups):
            layout = get_virtual_layout(fnames, gname)
            fout.create_virtual_dataset(gname, layout)

            # Copy first-file attributes to VDS root object
            with h5py.File(fnames[0], "r") as f0:
                for k, v in f0[gname].attrs.items():
                    fout[gname].attrs[k] = v

        # Build the cutBookkeeper
        counts_total = aggregate_cutbookkeeper(fnames=fnames, group_name=bookkeeper_name)
        if counts_total is not None:
            for sg, record in counts_total.items():
                grp = fout.require_group(f"{bookkeeper_name}/{sg}")
                grp.create_dataset("counts", data=record, shape=(), dtype=record.dtype)

    return out_fname


def main(args: Any | None = None) -> None:
    """Run VDS creation.

    Parameters
    ----------
    args : Any | None, optional
        Command line arguments, by default None
    """
    args = parse_args(args)
    matching_mode = "Applying regex to" if args.use_regex else "Globbing"
    print(f"{matching_mode} {args.pattern} ...")
    out_path = create_virtual_file(
        pattern=args.pattern,
        out_fname=args.output,
        use_regex=args.use_regex,
        regex_path=args.regex_path,
        overwrite=True,
    )

    with h5py.File(out_path, "r") as f:
        key = next(iter(f.keys()))
        print(f"Virtual dataset '{key}' has {len(f[key]):,} entries")

    print(f"Saved virtual file to {out_path.resolve()}")


if __name__ == "__main__":
    main()



from __future__ import annotations

from pathlib import Path
from tempfile import NamedTemporaryFile, mkdtemp

import h5py
import numpy as np
from numpy.lib.recfunctions import unstructured_to_structured as u2s

from ftag.hdf5 import join_structured_arrays

__all__ = ["get_mock_file"]

JET_VARS = [
    ("pt", "f4"),
    ("eta", "f4"),
    ("abs_eta", "f4"),
    ("mass", "f4"),
    ("pt_btagJes", "f4"),
    ("eta_btagJes", "f4"),
    ("n_tracks", "i4"),
    ("HadronConeExclTruthLabelID", "i4"),
    ("HadronConeExclTruthLabelPt", "f4"),
    ("R10TruthLabel_R22v1", "i4"),
    ("GhostBHadronsFinalCount", "i4"),
    ("GhostCHadronsFinalCount", "i4"),
    ("n_truth_promptLepton", "i4"),
    ("flavour_label", "i4"),
]

TRACK_VARS = [
    ("d0", "f4"),
    ("z0SinTheta", "f4"),
    ("dphi", "f4"),
    ("deta", "f4"),
    ("qOverP", "f4"),
    ("IP3D_signed_d0_significance", "f4"),
    ("IP3D_signed_z0_significance", "f4"),
    ("phiUncertainty", "f4"),
    ("thetaUncertainty", "f4"),
    ("qOverPUncertainty", "f4"),
    ("numberOfPixelHits", "u1"),
    ("numberOfSCTHits", "u1"),
    ("numberOfInnermostPixelLayerHits", "u1"),
    ("numberOfNextToInnermostPixelLayerHits", "u1"),
    ("numberOfInnermostPixelLayerSharedHits", "u1"),
    ("numberOfInnermostPixelLayerSplitHits", "u1"),
    ("numberOfPixelSharedHits", "u1"),
    ("numberOfPixelSplitHits", "u1"),
    ("numberOfSCTSharedHits", "u1"),
    ("numberOfPixelHoles", "u1"),
    ("numberOfSCTHoles", "u1"),
    ("leptonID", "i1"),
]


def softmax(x: np.ndarray, axis: int | None = None) -> np.ndarray:
    """Softmax function for numpy arrays.

    Parameters
    ----------
    x : np.ndarray
        Input array for the softmax
    axis : int | None, optional
        Axis along which the softmax is calculated, by default None

    Returns
    -------
    np.ndarray
        Output array with the softmax output
    """
    e_x = np.exp(x - np.max(x, axis=axis, keepdims=True))
    return e_x / e_x.sum(axis=axis, keepdims=True)


def get_mock_scores(labels: np.ndarray, is_xbb: bool = False) -> np.ndarray:
    if not is_xbb:
        label_dict = {"u": 0, "c": 4, "b": 5, "tau": 15}

    else:
        label_dict = {
            "hbb": 11,
            "hcc": 12,
            "top": 1,
            "qcd": 10,
            "htauel": 14,
            "htaumu": 15,
            "htauhad": 16,
        }

    # Set random seed
    rng = np.random.default_rng(42)

    # Set a list of possible means/scales
    mean_scale_list = [1, 2, 2.5, 3.5]

    # Get the number of classes
    n_classes = len(label_dict)

    # Init a scores array
    scores = np.zeros((len(labels), n_classes))

    # Generate means/scales
    means = []
    scales = []
    for i in range(n_classes):
        tmp_means = []
        tmp_means = [
            0 if j != i else mean_scale_list[rng.integers(0, len(mean_scale_list))]
            for j in range(n_classes)
        ]
        means.append(tmp_means)
        scales.append(mean_scale_list[rng.integers(0, len(mean_scale_list))])

    # Map the labels to the means
    label_mapping = dict(zip(label_dict.values(), means, strict=False))

    # Generate random mock scores
    for i, (label, count) in enumerate(zip(*np.unique(labels, return_counts=True), strict=False)):
        scores[labels == label] = rng.normal(
            loc=label_mapping[label], scale=scales[i], size=(count, n_classes)
        )

    # Pipe scores through softmax
    scores = softmax(scores, axis=1)
    name = "MockXbbTagger" if is_xbb else "MockTagger"
    cols = [f"{name}_p{x}" for x in label_dict]
    return u2s(scores, dtype=np.dtype([(name, "f4") for name in cols]))


def mock_jets(num_jets=1000) -> np.ndarray:
    # setup jets
    rng = np.random.default_rng(42)
    jets_dtype = np.dtype(JET_VARS)
    jets = u2s(rng.random((num_jets, len(JET_VARS))), jets_dtype)
    jets["flavour_label"] = rng.choice([0, 4, 5], size=num_jets)
    jets["pt"] *= 400e3
    jets["mass"] *= 50e3
    jets["eta"] = (jets["eta"] - 0.5) * 6.0
    jets["abs_eta"] = np.abs(jets["eta"])
    jets["n_truth_promptLepton"] = 0

    # add tagger scores
    jets["HadronConeExclTruthLabelID"] = rng.choice([0, 4, 5, 15], size=num_jets)
    jets["GhostBHadronsFinalCount"] = rng.choice([0, 1, 2], size=num_jets)
    jets["GhostCHadronsFinalCount"] = rng.choice([0, 1, 2], size=num_jets)
    jets["R10TruthLabel_R22v1"] = rng.choice([1, 10, 11, 12, 14, 15, 16], size=num_jets)
    scores = get_mock_scores(jets["HadronConeExclTruthLabelID"])
    xbb_scores = get_mock_scores(jets["R10TruthLabel_R22v1"], is_xbb=True)
    return join_structured_arrays([jets, scores, xbb_scores])


def mock_tracks(num_jets=1000, num_tracks=40) -> np.ndarray:
    rng = np.random.default_rng(42)
    tracks_dtype = np.dtype(TRACK_VARS)
    tracks = u2s(rng.random((num_jets, num_tracks, len(TRACK_VARS))), tracks_dtype)
    tracks["d0"] *= 5

    # for the shared hits, add some reasonable integer values
    tracks["numberOfPixelSharedHits"] = rng.integers(0, 3, size=(num_jets, num_tracks))
    tracks["numberOfSCTSharedHits"] = rng.integers(0, 3, size=(num_jets, num_tracks))

    valid = rng.choice([True, False], size=(num_jets, num_tracks))
    valid = valid.astype(bool).view(dtype=np.dtype([("valid", bool)]))
    return join_structured_arrays([tracks, valid])


def get_mock_file(
    num_jets: int = 1000,
    fname: str | None = None,
    tracks_name: str = "tracks",
    num_tracks: int = 40,
) -> tuple[str, h5py.File]:
    """Get a mock file for testing.

    Parameters
    ----------
    num_jets : int, optional
        Number of jets in the file, by default 1000
    fname : str | None, optional
        Name of the file, by default None
    tracks_name : str, optional
        Name of the tracks dataset, by default "tracks"
    num_tracks : int, optional
        Number of tracks per jet, by default 40

    Returns
    -------
    tuple[str, h5py.File]
        Tuple with the path and the h5 file
    """
    jets = mock_jets(num_jets)

    # create a tempfile in a new folder
    if fname is None:
        fname = NamedTemporaryFile(suffix=".h5", dir=mkdtemp()).name
    else:
        Path(fname).parent.mkdir(exist_ok=True, parents=True)
    f = h5py.File(fname, "w")
    f.create_dataset("jets", data=jets)
    f.attrs["test"] = "test"
    f["jets"].attrs["test"] = "test"

    # setup tracks
    if tracks_name:
        tracks = mock_tracks(num_jets, num_tracks)
        f.create_dataset(tracks_name, data=tracks)
    # Add some dummy meta-data
    counts_dtype = np.dtype([("count", "i8"), ("sum", "f8"), ("sum2", "f8")])
    counts_data = np.array([100, 200, 300], dtype=counts_dtype)
    f.create_dataset("cutBookkeeper/nominal/counts", data=counts_data)
    # add a test attribute to this group
    f["cutBookkeeper/nominal"].attrs["test"] = "test"
    return fname, f



from __future__ import annotations

from dataclasses import dataclass
from pathlib import Path

import h5py
import numpy as np

import ftag
from ftag.hdf5.h5utils import extract_group_full, write_group_full


@dataclass
class H5Writer:
    """Writes jets to an HDF5 file.

    Attributes
    ----------
    dst : Path | str
        Path to the output file.
    dtypes : dict[str, np.dtype]
        Dictionary of group names and their corresponding dtypes.
    shapes : dict[str, tuple[int, ...]], optional
        Dictionary of group names and their corresponding shapes.
    jets_name : str, optional
        Name of the jets group. Default is "jets".
    add_flavour_label : bool, optional
        Whether to add a flavour label to the jets group. Default is False.
    compression : str, optional
        Compression algorithm to use. Default is "lzf".
    precision : str, optional
        Precision to use. Default is None.
    full_precision_vars : list[str] | None, optional
        List of variables to store in full precision. Default is None.
    shuffle : bool, optional
        Whether to shuffle the jets before writing. Default is True.
    num_jets : int | None
        Number of jets to write.
    groups: dict[str, h5py.Group] | None
        Groups to copy from the source file. Default is None.
    """

    dst: Path | str
    dtypes: dict[str, np.dtype]
    shapes: dict[str, tuple[int, ...]]
    jets_name: str = "jets"
    add_flavour_label: bool = False
    compression: str = "lzf"
    precision: str = "full"
    full_precision_vars: list[str] | None = None
    shuffle: bool = True
    num_jets: int | None = None  # Allow dynamic mode by defaulting to None
    groups: dict[str, h5py.Group] | None = None  # Groups to copy from source file

    def __post_init__(self):
        self.num_written = 0
        self.rng = np.random.default_rng(42)

        # Infer number of jets from shapes if not explicitly passed
        inferred_num_jets = [shape[0] for shape in self.shapes.values()]
        if self.num_jets is None:
            assert len(set(inferred_num_jets)) == 1, "Shapes must agree in first dimension"
            self.fixed_mode = False
        else:
            self.fixed_mode = True
            for name in self.shapes:
                self.shapes[name] = (self.num_jets, *self.shapes[name][1:])

        if self.precision == "full":
            self.fp_dtype = np.float32
        elif self.precision == "half":
            self.fp_dtype = np.float16
        elif self.precision is None:
            self.fp_dtype = None
        else:
            raise ValueError(f"Invalid precision: {self.precision}")

        self.dst = Path(self.dst)
        self.dst.parent.mkdir(parents=True, exist_ok=True)
        self.file = h5py.File(self.dst, "w")
        self.add_attr("writer_version", ftag.__version__)

        for name, dtype in self.dtypes.items():
            self.create_ds(name, dtype)
        if self.groups:
            self.save_groups(self.groups)

    @classmethod
    def from_file(
        cls, source: Path, num_jets: int | None = 0, variables=None, copy_groups=True, **kwargs
    ) -> H5Writer:
        with h5py.File(source, "r") as f:
            dtypes = {}
            shapes = {}
            compression = []
            groups = {}
            for name, ds in f.items():
                if isinstance(ds, h5py.Group):
                    if copy_groups:
                        groups[name] = extract_group_full(ds)
                    continue
                if not isinstance(ds, h5py.Dataset):
                    raise TypeError(
                        f"Unsupported type {type(ds)} for dataset {name} in file {source}"
                    )
                dtypes[name] = ds.dtype
                shapes[name] = ds.shape
                compression.append(ds.compression)

            if variables:
                new_dtye = {}
                new_shape = {}
                for name, ds in f.items():
                    if name not in variables:
                        continue
                    new_dtye[name] = ftag.hdf5.get_dtype(
                        ds,
                        variables=variables[name],
                        precision=kwargs.get("precision"),
                        full_precision_vars=kwargs.get("full_precision_vars"),
                    )
                    new_shape[name] = ds.shape
                dtypes = new_dtye
                shapes = new_shape
            if num_jets != 0:
                shapes = {name: (num_jets, *shape[1:]) for name, shape in shapes.items()}

            assert len(set(compression)) == 1, "Must have same compression for all groups"
            compression = compression[0]
            if "compression" not in kwargs:
                kwargs["compression"] = compression
        return cls(dtypes=dtypes, shapes=shapes, groups=groups, **kwargs)

    def save_groups(self, groups: dict[str, dict]) -> None:
        for name, group_data in groups.items():
            if name not in self.file:
                write_group_full(self.file.create_group(name), group_data)

    def create_ds(self, name: str, dtype: np.dtype) -> None:
        if name == self.jets_name and self.add_flavour_label and "flavour_label" not in dtype.names:
            dtype = np.dtype([*dtype.descr, ("flavour_label", "i4")])

        fp_vars = self.full_precision_vars or []
        # If no precision is defined, or the field is in full_precision_vars, or its non-float,
        # keep it at the original dtype
        dtype = np.dtype([
            (
                field,
                (
                    self.fp_dtype
                    if (self.fp_dtype and field not in fp_vars and np.issubdtype(dt, np.floating))
                    else dt
                ),
            )
            for field, dt in dtype.descr
        ])

        shape = self.shapes[name]
        chunks = (100, *shape[1:]) if shape[1:] else None

        if self.fixed_mode:
            self.file.create_dataset(
                name, dtype=dtype, shape=shape, compression=self.compression, chunks=chunks
            )
        else:
            maxshape = (None, *shape[1:])
            self.file.create_dataset(
                name,
                dtype=dtype,
                shape=(0, *shape[1:]),
                maxshape=maxshape,
                compression=self.compression,
                chunks=chunks,
            )

    def close(self) -> None:
        if self.fixed_mode:
            written = len(self.file[self.jets_name])
            if self.num_written != written:
                raise ValueError(
                    f"Attempted to close file {self.dst} when only {self.num_written:,} out of"
                    f" {written:,} jets have been written"
                )
        self.file.close()

    def get_attr(self, name, group=None):
        with h5py.File(self.dst) as f:
            obj = f[group] if group else f
            return obj.attrs[name]

    def add_attr(self, name, data, group=None) -> None:
        obj = self.file[group] if group else self.file
        obj.attrs.create(name, data)

    def copy_attrs(self, fname: Path) -> None:
        with h5py.File(fname) as f:
            for name, value in f.attrs.items():
                self.add_attr(name, value)
            for name, ds in f.items():
                for attr_name, value in ds.attrs.items():
                    self.add_attr(attr_name, value, group=name)

    def write(self, data: dict[str, np.ndarray]) -> None:
        batch_size = len(data[self.jets_name])
        idx = np.arange(batch_size)
        if self.shuffle:
            self.rng.shuffle(idx)
            data = {name: array[idx] for name, array in data.items()}

        low = self.num_written
        high = low + batch_size

        if self.fixed_mode and high > self.num_jets:
            raise ValueError(
                f"Attempted to write more jets than expected: {high:,} > {self.num_jets:,}"
            )

        for group in self.dtypes:
            ds = self.file[group]
            if not self.fixed_mode:
                ds.resize((high, *ds.shape[1:]))
            ds[low:high] = data[group]

        self.num_written += batch_size



# Utils to take an input h5 file, and append one or more columns to it
from __future__ import annotations

import argparse
import importlib.util
from collections.abc import Callable
from pathlib import Path
from typing import Any

import h5py
import numpy as np

from ftag.hdf5.h5reader import H5Reader
from ftag.hdf5.h5writer import H5Writer


def merge_dicts(dicts: list[dict[str, dict[str, np.ndarray]]]) -> dict[str, dict[str, np.ndarray]]:
    """Merges a list of dictionaries.

    Each dict is of the form:
     {
        group1: {
            variable_1: np.array
            variable_2: np.array
        },
        group2: {
            variable_1: np.array
            variable_2: np.array
        }
     }

     E.g.

     dict1 = {
        "jets": {
            "pt": np.array([1, 2, 3]),
            "eta": np.array([4, 5, 6])
        },
    }
    dict2 = {
        "jets": {
            "phi": np.array([7, 8, 9]),
            "energy": np.array([10, 11, 12])
        },
    }

    merged = {
        "jets": {
            "pt": np.array([1, 2, 3]),
            "eta": np.array([4, 5, 6]),
            "phi": np.array([7, 8, 9]),
            "energy": np.array([10, 11, 12])
        }
    }

    Parameters
    ----------
    dicts : list[dict[str, dict[str, np.ndarray]]]
        List of dictionaries to merge. Each dictionary should be of the form:

    Returns
    -------
    dict[str, dict[str, np.ndarray]]
        Merged dictionary of the form:
        {
            group1: {
                variable_1: np.array
                variable_2: np.array
            },
            group2: {
                variable_1: np.array
                variable_2: np.array
            }
        }

    Raises
    ------
    ValueError
        If a variable already exists in the merged dictionary.
    """
    merged: dict[str, dict[str, np.ndarray]] = {}
    for d in dicts:
        for group, variables in d.items():
            if group not in merged:
                merged[group] = {}
            for variable, data in variables.items():
                if variable not in merged[group]:
                    merged[group][variable] = data
                else:
                    raise ValueError(f"Variable {variable} already exists in group {group}.")
    return merged


def get_shape(num_jets: int, batch: dict[str, np.ndarray]) -> dict[str, tuple[int, ...]]:
    """Returns a dictionary with the correct output shapes for the H5Writer.

    Parameters
    ----------
    num_jets : int
        Number of jets to write in total
    batch : dict[str, np.ndarray]
        Dictionary representing the batch

    Returns
    -------
    dict[str, tuple[int, ...]]
        Dictionary with the shapes of the output arrays
    """
    shape: dict[str, tuple[int, ...]] = {}

    for key, values in batch.items():
        if values.ndim == 1:
            shape[key] = (num_jets,)
        else:
            shape[key] = (num_jets, *values.shape[1:])
    return shape


def get_all_datasets(file: Path | str) -> dict[str, None]:
    """Returns a dictionary with all the groups in the h5 file.

    Parameters
    ----------
    file : Path | str
        Path to the h5 file

    Returns
    -------
    dict[str, None]
        A dictionary with all the datasets in the h5 file as keys and None as values,
        such that h5read.stream(all_groups) will return all the groups in the file.
    """
    with h5py.File(file, "r") as f:
        datasets = [d for d in f if isinstance(f[d], h5py.Dataset)]
        return dict.fromkeys(datasets)


def h5_add_column(
    input_file: str | Path,
    output_file: str | Path,
    append_function: Callable | list[Callable],
    num_jets: int = -1,
    input_groups: list[str] | None = None,
    output_groups: list[str] | None = None,
    reader_kwargs: dict | None = None,
    writer_kwargs: dict | None = None,
    overwrite: bool = False,
) -> None:
    """Appends one or more columns to one or more groups in an h5 file.

    Parameters
    ----------
    input_file : str | Path
        Input h5 file to read from.
    output_file : str | Path
        Output h5 file to write to.
    append_function : Callable | list[Callable]
        A function, or list of functions, which take a batch from H5Reader and returns a dictionary
        of the form:
            {
                group1 : {
                    new_column1 : data,
                    new_column2 : data,
                },
                group2 : {
                    new_column3 : data,
                    new_column4 : data,
                },
                ...
            }
    num_jets : int, optional
        Number of jets to read from the input file. If -1, reads all jets. By default -1.
    input_groups : list[str] | None, optional
        List of groups to read from the input file. If None, reads all groups. By default None.
    output_groups : list[str] | None, optional
        List of groups to write to the output file. If None, writes all groups. By default None.
        Note that this is a subset of the input groups, and must include all groups that the
        append functions wish to write to.
    reader_kwargs : dict | None, optional
        Additional arguments to pass to the H5Reader. By default None.
    writer_kwargs : dict | None, optional
        Additional arguments to pass to the H5Writer. By default None.
    overwrite : bool, optional
        If True, will overwrite the output file if it exists. By default False.
        If False, will raise a FileExistsError if the output file exists.
        If None, will check if the output file exists and raise an error if it does unless
        overwrite is True.

    Raises
    ------
    FileNotFoundError
        If the input file does not exist.
    FileExistsError
        If the output file exists and overwrite is False.
    ValueError
        If the new variable already exists, shape is incorrect, or the output group is not in
        the input groups.

    """
    input_file = Path(input_file)
    output_file = Path(output_file) if output_file is not None else None

    if not input_file.exists():
        raise FileNotFoundError(f"Input file {input_file} does not exist.")
    if output_file is not None and output_file.exists() and not overwrite:
        raise FileExistsError(
            f"Output file {output_file} already exists. Please choose a different name."
        )
    if not reader_kwargs:
        reader_kwargs = {}
    if not writer_kwargs:
        writer_kwargs = {}
    if output_file is None:
        output_file = input_file.with_name(input_file.name.replace(".h5", "_additional.h5"))

    if not isinstance(append_function, list):
        append_function = [append_function]

    reader = H5Reader(input_file, shuffle=False, **reader_kwargs)
    if "precision" not in writer_kwargs:
        writer_kwargs["precision"] = "full"

    njets = reader.num_jets if num_jets == -1 else num_jets
    writer = None

    input_variables = (
        get_all_datasets(input_file) if input_groups is None else dict.fromkeys(input_groups)
    )
    if output_groups is None:
        output_groups = list(input_variables.keys())

    assert all(o in input_variables for o in output_groups), (
        f"Output groups {output_groups} not in input groups {input_variables.keys()}"
    )

    num_batches = njets // reader.batch_size + 1
    for i, batch in enumerate(reader.stream(input_variables, num_jets=njets)):
        if (i + 1) % 10 == 0:
            print(f"Processing batch {i + 1}/{num_batches} ({(i + 1) / num_batches * 100:.2f}%)")

        to_append = merge_dicts([af(batch) for af in append_function])
        for k, newvars in to_append.items():
            if k not in output_groups:
                raise ValueError(f"Trying to output to {k} but only {output_groups} are allowed")
            for newkey, newval in newvars.items():
                if newkey in batch[k].dtype.names:
                    raise ValueError(
                        f"Trying to append {newkey} to {k} but it already exists in batch"
                    )
                if newval.shape != batch[k].shape:
                    raise ValueError(
                        f"Trying to append {newkey} to {k} but the shape is not correct"
                    )

        to_write = {}

        for key, str_array in batch.items():
            if key not in output_groups:
                continue
            if key in to_append:
                combined = np.lib.recfunctions.append_fields(
                    str_array,
                    list(to_append[key].keys()),
                    list(to_append[key].values()),
                    usemask=False,
                )
                to_write[key] = combined
            else:
                to_write[key] = str_array
        if writer is None:
            writer = H5Writer(
                output_file,
                dtypes={key: str_array.dtype for key, str_array in to_write.items()},
                shapes=get_shape(njets, to_write),
                shuffle=False,
                **writer_kwargs,
            )

        writer.write(to_write)


def parse_append_function(func_path: str) -> Callable:
    """Attempts to load the function specified by func_path.
    The function should be specified as 'path/to/file.py:function_name'.

    Parameters
    ----------
    func_path : str
        Path to the function to load. Should be of the form 'path/to/file.py:function_name'.

    Returns
    -------
    Callable
        The function specified by func_path.

    Raises
    ------
    ValueError
        If the function path is not of the form 'path/to/file.py:function_name'.
    FileNotFoundError
        If the file does not exist.
    ImportError
        If the file cannot be imported.
    AttributeError
        If the function does not exist in the file.
    """
    if isinstance(func_path, Path):
        func_path = str(func_path)
    if ":" not in func_path:
        print(func_path)
        raise ValueError("Function should be specified as 'path/to/file.py:function_name'")

    file_str, func_name = func_path.split(":")
    file_path = Path(file_str).resolve()

    if not file_path.is_file():
        raise FileNotFoundError(f"No such file: {file_path}")

    module_name = file_path.stem  # Just the filename without extension

    spec = importlib.util.spec_from_file_location(module_name, str(file_path))
    if spec is None or spec.loader is None:
        raise ImportError(f"Cannot load spec for {file_path}")

    module = importlib.util.module_from_spec(spec)
    spec.loader.exec_module(module)

    if not hasattr(module, func_name):
        raise AttributeError(f"Module {module_name} has no attribute {func_name}")

    return getattr(module, func_name)


def parse_args(args: Any | None) -> argparse.Namespace:
    """Parse command line arguments.

    Parameters
    ----------
    args : Any | None
        Command line arguments

    Returns
    -------
    argparse.Namespace
        Namespace with the parsed command line arguments.
    """
    parser = argparse.ArgumentParser(description="Append columns to an h5 file.")
    parser.add_argument("--input", "-i", type=str, required=True, help="Input h5 file")
    parser.add_argument(
        "--append_function",
        type=str,
        nargs="+",
        help="Function to append to the h5 file. Can be a list of functions.",
        required=True,
    )
    parser.add_argument("--output", type=str, help="Output h5 file")
    parser.add_argument(
        "--num_jets", type=int, default=-1, help="Number of jets to read from the input file"
    )
    parser.add_argument(
        "--input_groups",
        type=str,
        nargs="+",
        default=None,
        help="List of groups to read from the input file",
    )
    parser.add_argument(
        "--output_groups",
        type=str,
        nargs="+",
        default=None,
        help="List of groups to write to the output file",
    )
    parser.add_argument(
        "--reader_kwargs",
        type=dict,  # type: ignore[arg-type]
        default=None,
        help="Additional arguments for H5Reader",
    )
    parser.add_argument(
        "--writer_kwargs",
        type=dict,  # type: ignore[arg-type]
        default=None,
        help="Additional arguments for H5Writer",
    )
    parser.add_argument(
        "--overwrite", action="store_true", help="Overwrite the output file if it exists"
    )

    return parser.parse_args(args)


def main(args: Any | None = None) -> None:
    """Run VDS creation.

    Parameters
    ----------
    args : Any | None, optional
        Command line arguments, by default None
    """
    args = parse_args(args)
    append_function = [
        parse_append_function(func_path) if isinstance(func_path, str) else func_path
        for func_path in args.append_function
    ]

    h5_add_column(
        args.input,
        args.output,
        append_function,
        num_jets=args.num_jets,
        input_groups=args.input_groups,
        output_groups=args.output_groups,
        reader_kwargs=args.reader_kwargs,
        writer_kwargs=args.writer_kwargs,
        overwrite=args.overwrite,
    )



from __future__ import annotations

import argparse
import json
import re
from pathlib import Path
from typing import Any
from urllib.parse import ParseResult, urlparse

import h5py
import requests
import yaml

"""
find_metadata.py

Injects metadata (cross_section_pb, genFiltEff, kfactor) into .h5 files produced by TDD.

Usage:
    python find_metadata.py <h5_file1> [<h5_file2> ...] [-m fallback.yaml]
"""

XSECDB_MAP: dict[str, str] = {
    "mc15": "PMGxsecDB_mc15.txt",
    "mc16": "PMGxsecDB_mc16.txt",
    "mc21": "PMGxsecDB_mc21.txt",
    "mc23": "PMGxsecDB_mc23.txt",
}

XSECDB_URL_BASE: str = "https://atlas-groupdata.web.cern.ch/atlas-groupdata/dev/PMGTools/"


def validate_url_scheme(url: str) -> ParseResult:
    """
    Validate the scheme of a given URL, ensuring it is http or https.

    Parameters
    ----------
    url : str
        URL string to validate.

    Returns
    -------
    ParseResult
        Parsed URL object.

    Raises
    ------
    ValueError
        If the URL scheme is not supported.
    """
    parsed = urlparse(url)
    if parsed.scheme not in {"http", "https"}:
        raise ValueError(f"Unsupported URL scheme: {parsed.scheme}")
    return parsed


def download_xsecdb_files() -> None:
    """Download the PMG xsecDB files from CERN if they are not present locally."""
    for filename in XSECDB_MAP.values():
        if not Path(filename).exists():
            url = XSECDB_URL_BASE + filename
            try:
                validate_url_scheme(url)
                print(f"Downloading {filename} from {url}")
                response = requests.get(url, timeout=10)
                response.raise_for_status()
                Path(filename).write_bytes(response.content)
                print(f"Downloaded: {filename}")
            except (requests.RequestException, ValueError) as e:
                print(f"ERROR: Failed to download {filename} - {e}")


def extract_taskid_from_filename(h5_path: Path) -> str | None:
    """Extract the BigPanDA Task ID (8-digit) from an HDF5 filename.

    Parameters
    ----------
    h5_path : Path
        Path object pointing to the .h5 file.

    Returns
    -------
    str | None
        The Task ID as a string if found, otherwise None.
    """
    m = re.search(r"\.(\d{8})\.", h5_path.name)
    return m.group(1) if m else None


def fetch_taskinfo_from_bigpanda(taskid: str) -> dict[str, Any] | None:
    """Fetch task information from BigPanDA for a given Task ID.

    Parameters
    ----------
    taskid : str
        BigPanDA task ID.

    Returns
    -------
    dict[str, Any] | None
        Task info as a dictionary if found, otherwise None.
    """
    url = f"https://bigpanda.cern.ch/tasks/?jeditaskid={taskid}&json"
    print(f"Fetching task info from: {url}")
    try:
        validate_url_scheme(url)
        response = requests.get(url, timeout=10)
        response.raise_for_status()
        data = response.json()
        if isinstance(data, list) and data:
            return data[0]
        print("No task info found.")
    except (requests.RequestException, ValueError, json.JSONDecodeError) as e:
        print(f"ERROR: Failed to fetch task info for {taskid} - {e}")
    return None


def extract_mc_container_from_json(data: dict[str, Any]) -> str | None:
    """Extract the MC container name (e.g., mc16_13TeV.<something>) from a task JSON.

    Parameters
    ----------
    data : dict[str, Any]
        Task info dictionary from BigPanDA.

    Returns
    -------
    str | None
        The container string if found, otherwise None.
    """
    text = json.dumps(data)
    match = re.search(r"(mc\d+_13TeV\.[\w\.]+)", text)
    return match.group(1) if match else None


def parse_line_from_taskname(taskname: str) -> tuple[int | None, str | None]:
    """Extract DSID and etag from a task name string.

    Parameters
    ----------
    taskname : str
        Full task name.

    Returns
    -------
    tuple[int | None, str | None]
        A tuple of (DSID as int, etag as string), or (None, None) if not found.
    """
    match = re.match(r".*\.(\d{6})\.e(\d+)_", taskname)
    if match:
        dsid, etag = match.groups()
        return int(dsid), f"e{etag}"
    return None, None


def parse_campaign_from_taskname(taskname: str) -> str | None:
    """Derive campaign (mc15/mc16/etc.) from a task or container name.

    Parameters
    ----------
    taskname : str
        The name string.

    Returns
    -------
    str | None
        Campaign string, or None if not found.
    """
    m = re.match(r"(mc\d+)_13TeV", taskname)
    if m:
        raw_campaign = m.group(1)
        return "mc16" if raw_campaign == "mc20" else raw_campaign
    return None


def extract_info_from_container(container: str) -> tuple[int, str, str] | None:
    """Extract DSID, etag, and campaign name from a container string.

    Parameters
    ----------
    container : str
        The MC container string.

    Returns
    -------
    tuple[int, str, str] | None
        A tuple of (DSID, etag, campaign), or None if parsing fails.
    """
    m_campaign = re.search(r"\b(mc\d+)_13TeV", container)
    m_dsid = re.search(r"\.(\d{6})\.", container)
    m_etag = re.search(r"\.e(\d+)(?:[_.]|$)", container)
    if m_campaign and m_dsid and m_etag:
        raw_campaign = m_campaign.group(1)
        campaign = "mc16" if raw_campaign == "mc20" else raw_campaign
        dsid = int(m_dsid.group(1))
        etag = f"e{m_etag.group(1)}"
        return dsid, etag, campaign
    return None


def query_xsecdb(campaign: str, dsid: int, etag: str) -> dict[str, Any] | None:
    """Look up cross-section metadata in the PMG xsecDB.

    Parameters
    ----------
    campaign : str
        Campaign name (e.g., mc16).
    dsid : int
        Dataset ID.
    etag : str
        Event tag.

    Returns
    -------
    dict[str, Any] | None
        Dictionary with cross_section_pb, genFiltEff, kfactor, and etag if found, otherwise None.
    """
    db_path = XSECDB_MAP.get(campaign)
    print(f"Searching in {db_path} for DSID={dsid}, etag={etag}")
    if not db_path or not Path(db_path).exists():
        print(f"ERROR: Database file for campaign '{campaign}' not found.")
        return None

    # Iterate through lines in the DB to find a match
    with open(db_path) as f:
        for raw_line in f:
            line = raw_line.strip()
            if not line or line.startswith("#"):
                continue
            fields = line.split()
            if len(fields) < 9:
                continue
            if str(dsid) == fields[0] and etag == fields[8]:
                print(f"Match found in {db_path}")
                return {
                    "cross_section_pb": float(fields[2]),
                    "genFiltEff": float(fields[3]),
                    "kfactor": float(fields[4]),
                    "etag": fields[8],
                }
    print(f"No match found for DSID={dsid}, etag={etag} in {db_path}")
    return None


def write_metadata_to_h5(h5_filename: str, dsid: int, metadata_dict: dict[str, Any]) -> None:
    """Write metadata values into an HDF5 file under metadata/<DSID>.

    Parameters
    ----------
    h5_filename : str
        Target HDF5 file.
    dsid : int
        Dataset ID to write metadata for.
    metadata_dict : dict[str, Any]
        Dictionary of metadata to inject.
    """
    with h5py.File(h5_filename, "a") as f:
        meta_group = f.require_group("metadata")
        dsid_group = meta_group.require_group(str(dsid))

        # Overwrite existing keys if present
        for key in ["cross_section_pb", "genFiltEff", "kfactor"]:
            if key in dsid_group:
                del dsid_group[key]
            dsid_group.create_dataset(key, data=metadata_dict[key])
            print(f"Wrote {key} = {metadata_dict[key]} to metadata/{dsid}/{key}")


def handle_yaml_fallback(h5_path: Path, yaml_data: dict[str, Any]) -> None:
    """Use fallback metadata from YAML if automatic lookup fails.

    Parameters
    ----------
    h5_path : Path
        Path to the HDF5 file.
    yaml_data : dict[str, Any]
        Metadata dictionary loaded from YAML.

    Raises
    ------
    ValueError
        If YAML is invalid, empty, or missing required fields.
    """
    if len(yaml_data) != 1:
        raise ValueError("YAML fallback file must contain exactly one entry")

    key, value = next(iter(yaml_data.items()))

    # YAML may specify container or DSID
    if key == "container name":
        if not isinstance(value, str):
            raise ValueError("'container name' must be a string")
        container = value
        print(f"YAML container fallback: {container}")
        info = extract_info_from_container(container)
        if not info:
            raise ValueError(
                f"Failed to parse valid DSID / etag / campaign from container: {container}"
            )
        dsid, etag, campaign = info
        meta = query_xsecdb(campaign, dsid, etag)
        if not meta:
            raise ValueError(f"Container not found in PMG database: {container}")
    else:
        try:
            dsid = int(key)
        except ValueError as e:
            raise ValueError("YAML key must be a valid DSID or 'container name'") from e
        meta = value
        required_keys = {"cross_section_pb", "genFiltEff", "kfactor"}
        if not isinstance(meta, dict) or not required_keys.issubset(meta):
            raise ValueError(
                "Manual metadata must include: cross_section_pb / genFiltEff / kfactor"
            )

    write_metadata_to_h5(str(h5_path), dsid, meta)
    print(f"YAML metadata written to {h5_path.name} for DSID {dsid}")


def parse_args_and_yaml() -> tuple[list[str], dict[str, Any]]:
    """Parse CLI arguments and load YAML metadata if provided.

    Returns
    -------
    tuple[list[str], dict[str, Any]]
        A tuple of (list of HDF5 file paths, YAML metadata dict).
    """
    parser = argparse.ArgumentParser(
        description="Inject metadata into .h5 files via BigPanDA or YAML fallback."
    )
    parser.add_argument("h5_files", nargs="+", help="List of .h5 files")
    parser.add_argument("-m", "--manual-yaml", help="Manual metadata YAML fallback file")
    args = parser.parse_args()

    yaml_data: dict[str, Any] = {}
    if args.manual_yaml:
        with open(args.manual_yaml) as f:
            yaml_data = yaml.safe_load(f)

    return args.h5_files, yaml_data


def process_single_file(path: Path, yaml_data: dict[str, Any]) -> None:
    """Process a single .h5 file by attempting BigPanDA lookup, then fallback to YAML.

    Parameters
    ----------
    path : Path
        Path to the HDF5 file.
    yaml_data : dict[str, Any]
        Optional fallback metadata.
    """
    if not path.exists():
        print(f"File not found: {path}")
        return

    taskid = extract_taskid_from_filename(path)
    if taskid:
        print(f"Extracted Task ID: {taskid}")
        taskinfo = fetch_taskinfo_from_bigpanda(taskid)
        if taskinfo:
            taskname = taskinfo.get("taskname", "")
            dsid, etag = parse_line_from_taskname(taskname)
            container = extract_mc_container_from_json(taskinfo)
            campaign = parse_campaign_from_taskname(container or "")

            # Only query DB if all three are found
            if dsid and etag and campaign:
                print(
                    f"Matched info: DSID={dsid}, etag={etag}, "
                    f"campaign={campaign}, container={container}"
                )
                meta = query_xsecdb(campaign, dsid, etag)
                if meta:
                    meta["campaign"] = campaign
                    print(
                        f"Got metadata: cross_section = {meta['cross_section_pb']} pb, "
                        f"eff = {meta['genFiltEff']}, k = {meta['kfactor']}"
                    )
                    try:
                        write_metadata_to_h5(str(path), dsid, meta)
                        print(f"Metadata written to {path.name}")
                    except (OSError, ValueError) as e:
                        print(f"Failed to write metadata: {e}")
                    else:
                        return
    else:
        print("Failed to extract Task ID from filename")

    # Fallback to YAML if automatic mode fails
    if yaml_data:
        try:
            print("Using YAML fallback path")
            handle_yaml_fallback(path, yaml_data)
        except (ValueError, KeyError, OSError) as e:
            print(f"YAML fallback failed: {e}")
    else:
        print("Failed to retrieve metadata and no YAML fallback provided")


def main() -> None:
    """Entry point: parse arguments, download xsecDBs, process each file, and clean up."""
    h5_files, yaml_data = parse_args_and_yaml()
    download_xsecdb_files()

    for h5_file in h5_files:
        print(f"\n==================== Processing {h5_file} ====================\n")
        process_single_file(Path(h5_file), yaml_data)

    # Remove temporary database files after use
    for filename in XSECDB_MAP.values():
        path = Path(filename)
        if path.exists():
            try:
                path.unlink()
                print(f"Deleted temporary file: {filename}")
            except OSError as e:
                print(f"Failed to delete {filename}: {e}")


if __name__ == "__main__":
    main()



"""Utils package of the atlas-ftag-tools."""

from __future__ import annotations

from .logging import logger, set_log_level
from .metrics import (
    calculate_efficiency,
    calculate_efficiency_error,
    calculate_rejection,
    calculate_rejection_error,
    get_discriminant,
    save_divide,
    weighted_percentile,
)

__all__ = [
    "calculate_efficiency",
    "calculate_efficiency_error",
    "calculate_rejection",
    "calculate_rejection_error",
    "get_discriminant",
    "logger",
    "save_divide",
    "set_log_level",
    "weighted_percentile",
]



"""Small utilities for interacting with a local Git repository.

This module provides helper functions to:

- detect whether a given path is inside a Git working tree,
- fail fast when there are uncommitted changes,
- verify that a local clone is a fork of an expected upstream,
- create and push an annotated tag, and
- read the current short commit hash.

All functions run Git commands via :mod:`subprocess`. If Git is not installed
or the command cannot be executed, the underlying OS error (e.g. :class:`OSError`)
will propagate.
"""

from __future__ import annotations

import subprocess
import sys
from os import PathLike
from subprocess import CalledProcessError


class GitError(Exception):
    """Raised when a Git-related precondition is not satisfied."""


def is_git_repo(path: str | PathLike[str]) -> bool:
    """Return whether ``path`` is inside a Git working tree.

    Parameters
    ----------
    path : str | PathLike[str]
        Filesystem path used as the current working directory for the Git command.

    Returns
    -------
    bool
        ``True`` if ``path`` is inside a Git working tree, ``False`` otherwise.

    Notes
    -----
    This function runs::

        git rev-parse --is-inside-work-tree HEAD

    Any non-zero exit status is treated as "not a Git repository". If Git is not
    available on the system, an :class:`OSError` may be raised by :mod:`subprocess`.
    """
    try:
        subprocess.check_output(
            ["git", "rev-parse", "--is-inside-work-tree", "HEAD"],
            cwd=path,
        )
    except CalledProcessError:
        return False
    else:
        return True


def check_for_uncommitted_changes(path: str | PathLike[str]) -> None:
    """Raise if the repository at ``path`` has uncommitted changes.

    Parameters
    ----------
    path : str | PathLike[str]
        Filesystem path to the repository root or any directory within it.

    Raises
    ------
    GitError
        If ``path`` is a Git repository and there are uncommitted changes.

    Notes
    -----
    - If ``path`` is **not** a Git repository, the function returns silently.
    - If the current process is running under ``pytest`` (detected via
      :data:`sys.modules`), the check is skipped and the function returns.
    """
    if not is_git_repo(path):
        return
    if "pytest" in sys.modules:
        return

    try:
        subprocess.check_output(
            ["git", "diff", "--quiet", "--exit-code"],
            cwd=path,
        )
    except CalledProcessError:
        raise GitError(
            "Uncommitted changes detected. Please commit them before running, or use --force."
        ) from None


def check_for_fork(path: str | PathLike[str], upstream: str) -> None:
    """Ensure the local clone's ``origin`` remote is a fork of ``upstream``.

    Parameters
    ----------
    path : str | PathLike[str]
        Filesystem path to the repository root or any directory within it.
    upstream : str
        Expected upstream repository URL substring (e.g. ``'github.com/org/repo'``).

    Raises
    ------
    GitError
        If the repository is present but its ``origin`` URL does not contain ``upstream``.

    Notes
    -----
    If ``path`` is not a Git repository, the function returns silently.
    """
    if not is_git_repo(path):
        return

    cmd = ["git", "remote", "get-url", "origin"]
    origin = subprocess.check_output(cmd, cwd=path).decode("utf-8").strip()
    if upstream not in origin:
        raise GitError(f"Your origin {origin} is not a fork of the upstream repo {upstream}")


def create_and_push_tag(
    path: str | PathLike[str],
    upstream: str,
    tagname: str,
    msg: str,
) -> None:
    """Create an annotated Git tag and push it to ``origin``.

    Parameters
    ----------
    path : str | PathLike[str]
        Filesystem path to the repository root or any directory within it.
    upstream : str
        Expected upstream repository URL substring; passed to :func:`check_for_fork`.
    tagname : str
        Name of the tag to create.
    msg : str
        Annotation message for the tag (``git tag -m``).

    Notes
    -----
    If ``path`` is not a Git repository, the function returns silently.
    """
    print(f"Pushing tag {tagname}")
    if not is_git_repo(path):
        return
    check_for_fork(path, upstream)
    subprocess.check_output(["git", "tag", tagname, "-m", msg], cwd=path)
    subprocess.check_output(["git", "push", "-q", "origin", "--tags"], cwd=path)


def get_git_hash(path: str | PathLike[str]) -> str | None:
    """Return the short commit hash for ``HEAD`` at ``path``, if available.

    Parameters
    ----------
    path : str | PathLike[str]
        Filesystem path to the repository root or any directory within it.

    Returns
    -------
    str | None
        The short (``--short``) commit hash as a string, or ``None`` if ``path`` is
        not a Git repository.
    """
    if not is_git_repo(path):
        return None

    git_hash = subprocess.check_output(
        ["git", "rev-parse", "--short", "HEAD"],
        cwd=path,
    )
    return git_hash.decode("ascii").strip()



from __future__ import annotations

import math
from collections.abc import Callable, Generator
from dataclasses import dataclass
from functools import cached_property
from pathlib import Path

import h5py
import numpy as np

from ftag.cuts import Cuts
from ftag.hdf5.h5utils import get_dtype
from ftag.sample import Sample
from ftag.transform import Transform
from ftag.utils.logging import logger


@dataclass
class H5SingleReader:
    """H5 single reader for a single file.

    Attributes
    ----------
    fname: Path | str
        Path to the file
    batch_size: int, optional
        Batch size to read, by default 100_000
    jets_name: str, optional
        Name of the jets dataset, by default "jets"
    precision: str | None, optional
        Precision that is to be used, by default None
    shuffle: bool, optional
        If random jets are loaded from the file, by default True
    do_remove_inf: bool, optional
        Remove infs from the jets, by default False
    transform: Transform | None, optional
        Transformation that should be applied, by default None
    groups: list[str] | None, optional
        List of the groups that hold metadata, by default None
    dsets: list[str] | None, optional
        List of the datsets, that hold the per jet/track data, by default None
    vds_dir: Path | str | None, optional
        Directory where virtual datasets will be stored if wildcard is used, by default None.
        If None, the virtual files will be created in the same directory as the input files.
    """

    fname: Path | str
    batch_size: int = 100_000
    jets_name: str = "jets"
    precision: str | None = None
    shuffle: bool = True
    do_remove_inf: bool = False
    transform: Transform | None = None
    # groups hold meta-data e.g. "cutBookkeeper"
    groups: list[str] | None = None
    # dsets hold data and all have a first dimension of njets
    dsets: list[str] | None = None
    vds_dir: Path | str | None = None

    def __post_init__(self) -> None:
        self.rng = np.random.default_rng(42)
        self.sample = Sample(self.fname, vds_dir=self.vds_dir)
        fname = self.sample.virtual_file()
        if len(fname) != 1:
            raise ValueError("H5SingleReader should only read a single file")
        self.fname = fname[0]
        with h5py.File(self.fname) as f:
            self.groups = self.groups or [g for g in f if isinstance(f[g], h5py.Group)]
            self.dsets = self.dsets or [d for d in f if isinstance(f[d], h5py.Dataset)]

    @cached_property
    def num_jets(self) -> int:
        with h5py.File(self.fname) as f:
            return len(f[self.jets_name])

    def get_attr(self, name, group=None):
        with h5py.File(self.fname) as f:
            obj = f[group] if group else f
            return obj.attrs[name]

    def empty(self, ds: h5py.Dataset | h5py.Group, variables: list[str]) -> np.ndarray:
        if variables is None:
            variables = ds.dtype.names

        return np.array(0, dtype=get_dtype(ds, variables, self.precision, transform=self.transform))

    def read_chunk(self, ds: h5py.Dataset, array: np.ndarray, low: int) -> np.ndarray:
        high = min(low + self.batch_size, self.num_jets)
        shape = (high - low, *ds.shape[1:])
        array.resize(shape, refcheck=False)
        ds.read_direct(array, np.s_[low:high])
        return array

    def remove_inf(self, data: dict) -> dict:
        keep_idx = np.full(len(data[self.jets_name]), True)
        for name, array in data.items():
            for var in array.dtype.names:
                isinf = np.isinf(array[var])
                isinf = isinf if name == self.jets_name else isinf.any(axis=-1)
                keep_idx &= ~isinf
                if num_inf := isinf.sum():
                    logger.warning(
                        f"{num_inf} inf values detected for variable {var} in"
                        f" {name} array. Removing the affected jets."
                    )
        return {name: array[keep_idx] for name, array in data.items()}

    def _process_batch(self, data: dict, cuts: Cuts | None = None) -> dict:
        """Apply cuts and transformations to the batch.

        Parameters
        ----------
        data : dict
            Dictionary of arrays for each group.
        cuts : Cuts | None, optional
            Selection cuts to apply, by default None

        Returns
        -------
        dict
            Processed data dictionary with arrays for each group. After applying cuts,
            (optional) removal of infs, and (optional) transformation.
        """
        # apply selections
        if cuts:
            idx = cuts(data[self.jets_name]).idx
            data = {name: array[idx] for name, array in data.items()}

        # check for inf and remove
        if self.do_remove_inf:
            data = self.remove_inf(data)

        # apply transform
        if self.transform:
            data = self.transform(data)

        return data

    def stream(
        self,
        variables: dict | None = None,
        num_jets: int | None = None,
        cuts: Cuts | None = None,
        start: int = 0,
        skip_batches: int = 0,
    ) -> Generator:
        if num_jets is None:
            num_jets = self.num_jets
        if skip_batches > 0:
            assert not self.shuffle, "Cannot skip batches if shuffle is True"
        if num_jets > self.num_jets:
            logger.warning(
                f"{num_jets:,} jets requested but only {self.num_jets:,} available in {self.fname}."
                " Set to maximum available number!"
            )
            num_jets = self.num_jets

        if variables is None:
            variables = {self.jets_name: None}

        total = 0
        with h5py.File(self.fname) as f:
            arrays = {name: self.empty(f[name], var) for name, var in variables.items()}
            data = {name: self.empty(f[name], var) for name, var in variables.items()}

            # get indices
            indices = list(range(start, self.num_jets + start, self.batch_size))
            if self.shuffle:
                self.rng.shuffle(indices)
            if skip_batches > 0:
                indices = indices[skip_batches:]
            # loop over batches and read file
            for low in indices:
                for name in variables:
                    data[name] = self.read_chunk(f[name], arrays[name], low)

                # Apply cuts and transformations
                data = self._process_batch(data, cuts)

                # check for completion
                total += len(data[self.jets_name])
                if total >= num_jets:
                    keep = num_jets - (total - len(data[self.jets_name]))
                    data = {name: array[:keep] for name, array in data.items()}
                    yield data
                    break

                yield data

    def get_batch_reader(
        self,
        variables: dict | None = None,
        cuts: Cuts | None = None,
    ) -> Callable:
        """Get a function to read batches of selected jets.

        Parameters
        ----------
        variables : dict | None, optional
            Dictionary of variables to for each group, by default use all jet variables.
        cuts : Cuts | None, optional
            Selection cuts to apply, by default None

        Returns
        -------
        Callable
            Function that takes an index and returns a batch of selected jets.
        """
        if variables is None:
            variables = {self.jets_name: None}
        h5 = h5py.File(self.fname, "r")
        arrays = {name: self.empty(h5[name], var) for name, var in variables.items()}
        # nonlocal data
        data = {name: self.empty(h5[name], var) for name, var in variables.items()}

        def get_batch(idx: int) -> dict | None:
            """Get a batch of data from the HDF5 file.

            Parameters
            ----------
            idx : int
                Index of the batch to read.

            Returns
            -------
            dict | None
                Dictionary of arrays for each group, or None if no more batches are available.
            """
            low = idx * self.batch_size
            if low >= self.num_jets:
                return None

            for name in variables:
                data[name] = self.read_chunk(h5[name], arrays[name], low)

            data_out = {name: array.copy() for name, array in data.items()}

            return self._process_batch(data_out, cuts)

        return get_batch


@dataclass
class H5Reader:
    """Reads data from multiple HDF5 files.

    Attributes
    ----------
    fname : Path | str | list[Path | str]
        Path to the HDF5 file or list of paths
    batch_size : int, optional
        Number of jets to read at a time, by default 100_000
    jets_name : str, optional
        Name of the jets dataset, by default "jets"
    precision : str | None, optional
        Cast floats to given precision, by default None
    shuffle : bool, optional
        Read batches in a shuffled order, by default True
    weights : list[float] | None, optional
        Weights for different input datasets, by default None
    do_remove_inf : bool, optional
        Remove jets with inf values, by default False
    transform : Transform | None, optional
        Transform to apply to data, by default None
    equal_jets : bool, optional
        Take the same number of jets (weighted) from each sample, by default True.
        This is useful when you specify a list of DSIDs for the sample and they are
        qualitatively different, and you want to ensure that you always return batches
        with jets from all DSIDs. This is used for example in the QCD resampling for Xbb.
        If False, use all jets in each sample, allowing for the full available statistics
        to be used. Useful for example if you have multiple ttbar samples and you want to
        use all available jets from each sample.
    vds_dir: Path | str | None, optional
        Directory where virtual datasets will be stored if wildcard is used, by default None.
        If None, the virtual files will be created in the same directory as the input files.
    """

    fname: Path | str | list[Path | str]
    batch_size: int = 100_000
    jets_name: str = "jets"
    precision: str | None = None
    shuffle: bool = True
    weights: list[float] | None = None
    do_remove_inf: bool = False
    transform: Transform | None = None
    equal_jets: bool = False
    vds_dir: Path | str | None = None

    def __post_init__(self) -> None:
        self.rng = np.random.default_rng(42)
        if isinstance(self.fname, str | Path):
            self.fname = [self.fname]

        # calculate batch sizes
        if self.weights is None:
            rows_per_file = [
                H5SingleReader(f, jets_name=self.jets_name, vds_dir=self.vds_dir).num_jets
                for f in self.fname
            ]
            num_total = sum(rows_per_file)
            self.weights = [num / num_total for num in rows_per_file]

        self.batch_sizes = [int(w * self.batch_size) for w in self.weights]

        # create readers
        self.readers = [
            H5SingleReader(
                f,
                b,
                self.jets_name,
                self.precision,
                self.shuffle,
                self.do_remove_inf,
                self.transform,
                vds_dir=self.vds_dir,
            )
            for f, b in zip(self.fname, self.batch_sizes, strict=False)
        ]

    @property
    def num_jets(self) -> int:
        return sum(r.num_jets for r in self.readers)

    @property
    def files(self) -> list[Path]:
        return [Path(r.fname) for r in self.readers]

    def dtypes(self, variables: dict[str, list[str]] | None = None) -> dict[str, np.dtype]:
        dtypes = {}
        with h5py.File(self.files[0]) as f:
            if variables is None:
                for key in f:
                    if isinstance(f[key], h5py.Group):
                        continue
                    if isinstance(f[key], h5py.Dataset):
                        dtype = f[key].dtype
                    if self.transform:
                        dtype = self.transform.map_dtype(key, dtype)
                    dtypes[key] = dtype
            else:
                for name, var in variables.items():
                    ds = f[name]
                    dtype = get_dtype(ds, var, self.precision, transform=self.transform)
                    dtypes[name] = dtype
        return dtypes

    def shapes(self, num_jets: int, groups: list[str] | None = None) -> dict[str, tuple[int, ...]]:
        if groups is None:
            groups = [self.jets_name]
        shapes = {}
        with h5py.File(self.files[0]) as f:
            for group in groups:
                shape = f[group].shape
                shapes[group] = (num_jets, *shape[1:])
        return shapes

    def stream(
        self,
        variables: dict | None = None,
        num_jets: int | None = None,
        cuts: Cuts | None = None,
        start: int = 0,
        skip_batches: int = 0,
    ) -> Generator:
        """Generate batches of selected jets.

        Parameters
        ----------
        variables : dict | None, optional
            Dictionary of variables to for each group, by default use all jet variables.
        num_jets : int | None, optional
            Total number of selected jets to generate, by default all.
        cuts : Cuts | None, optional
            Selection cuts to apply, by default None
        start : int, optional
            Starting index of the first jet to read, by default 0
        skip_batches : int, optional
            Number of batches to skip, by default 0

        Yields
        ------
        Generator
            Generator of batches of selected jets.
        """
        # Check if number of jets is given, if not, set to maximum available
        if num_jets is None:
            num_jets = self.num_jets

        # Check if variables if given, if not, set to all
        if variables is None:
            variables = {self.jets_name: None}

        if self.jets_name not in variables or variables[self.jets_name] is not None:
            jet_vars = variables.get(self.jets_name, [])
            variables[self.jets_name] = list(jet_vars) + (cuts.variables if cuts else [])

        # get streams for selected jets from each reader
        streams = [
            r.stream(
                variables, int(r.num_jets / self.num_jets * num_jets), cuts, start, skip_batches
            )
            for r in self.readers
        ]

        while True:
            samples = []
            # Track which streams have been exhausted
            streams_done = [False] * len(streams)

            # for each unexhausted stream, get the next sample
            for i, stream in enumerate(streams):
                if not streams_done[i]:
                    try:
                        samples.append(next(stream))

                    # if equal_jets is True, stop when any sample is done
                    # otherwise if stream is exhausted, mark it as such and continue
                    except StopIteration:
                        if self.equal_jets:
                            return
                        streams_done[i] = True

                # if equal_jets is False, we need to keep going until all streams are done
                if all(streams_done):
                    return

            # combine samples and shuffle
            data = {name: np.concatenate([s[name] for s in samples]) for name in variables}
            if self.shuffle:
                idx = np.arange(len(data[self.jets_name]))
                self.rng.shuffle(idx)
                data = {name: array[idx] for name, array in data.items()}

            # yield batch
            yield data

    def get_batch_reader(
        self,
        variables: dict | None = None,
        cuts: Cuts | None = None,
        shuffle: bool = True,
    ) -> Callable:
        """Get a function to read batches of selected jets.

        Parameters
        ----------
        variables : dict | None, optional
            Dictionary of variables to for each group, by default use all jet variables.
        cuts : Cuts | None, optional
            Selection cuts to apply, by default None
        shuffle : bool, optional
            Read batches in a shuffled order, by default True

        Returns
        -------
        Callable
            Function that takes an index and returns a batch of selected jets.
        """
        if variables is None:
            variables = {self.jets_name: None}

        # create batch readers for each sample
        batch_readers = [r.get_batch_reader(variables, cuts) for r in self.readers]

        def get_batch(idx: int) -> dict | None:
            """Get a batch of data from the HDF5 files.

            Parameters
            ----------
            idx : int
                Index of the batch to read.

            Returns
            -------
            dict | None
                Dictionary of arrays for each group, or None if no more batches are available.
            """
            assert idx >= 0, "Index must be non-negative"
            if idx * self.batch_size >= self.num_jets:
                return None
            # get a batch from each sample
            samples = [br(idx) for br in batch_readers]
            samples = [s for s in samples if s is not None]
            if len(samples) == 0:
                return None
            # combine samples and shuffle
            data = {name: np.concatenate([s[name] for s in samples]) for name in variables}
            if shuffle:
                idx = np.arange(len(data[self.jets_name]))
                self.rng.shuffle(idx)
                data = {name: array[idx] for name, array in data.items()}
            return data

        return get_batch

    def load(
        self, variables: dict | None = None, num_jets: int | None = None, cuts: Cuts | None = None
    ) -> dict:
        """Load multiple batches of selected jets into memory.

        Parameters
        ----------
        variables : dict | None, optional
            Dictionary of variables to for each group, by default use all jet variables.
        num_jets : int | None, optional
            Total number of selected jets to load, by default all.
        cuts : Cuts | None, optional
            Selection cuts to apply, by default None

        Returns
        -------
        dict
            Dictionary of arrays for each group.
        """
        # handle default arguments
        if num_jets == -1:
            num_jets = self.num_jets
        if variables is None:
            variables = {self.jets_name: None}

        # get data from each sample
        data: dict[str, list] = {name: [] for name in variables}
        for batch in self.stream(variables, num_jets, cuts):
            for name, array in batch.items():
                if name in data:
                    data[name].append(array)

        # concatenate batches
        return {name: np.concatenate(array) for name, array in data.items()}

    def estimate_available_jets(self, cuts: Cuts, num: int = 1_000_000) -> int:
        """Estimate the number of jets available after selection cuts.

        Parameters
        ----------
        cuts : Cuts
            Selection cuts to apply.
        num : int, optional
            Number of jets to use for the estimation, by default 1_000_000.

        Returns
        -------
        int
            Estimated number of jets available after selection cuts, rounded down.
        """
        # reset rngs to ensure same jets are used for each sample
        self.rng = np.random.default_rng(42)
        for r in self.readers:
            r.rng = np.random.default_rng(42)

        # if equal jets is True, available jets is based on the smallest sample
        if self.equal_jets:
            num_jets = []
            for r in self.readers:
                stream = r.stream({self.jets_name: cuts.variables}, num)
                all_jets = np.concatenate([batch[self.jets_name].copy() for batch in stream])
                frac_selected = len(cuts(all_jets).values) / len(all_jets)
                num_jets.append(frac_selected * r.num_jets)
            estimated_num_jets = min(num_jets) * len(self.readers)
        # otherwise, available jets is based on all samples
        else:
            all_jets = self.load({self.jets_name: cuts.variables}, num)[self.jets_name]
            frac_selected = len(cuts(all_jets).values) / len(all_jets)
            estimated_num_jets = frac_selected * self.num_jets
        return math.floor(estimated_num_jets * 0.99)



"""atlas-ftag-tools - Common tools for ATLAS flavour tagging software."""

from __future__ import annotations

__version__ = "v0.3.2"

from . import hdf5, utils
from .cuts import Cuts
from .flavours import Extended_Flavours, Flavours
from .fraction_optimization import calculate_best_fraction_values
from .git_check import (
    GitError,
    check_for_fork,
    check_for_uncommitted_changes,
    create_and_push_tag,
    get_git_hash,
    is_git_repo,
)
from .labeller import Labeller
from .labels import Label, LabelContainer
from .mock import get_mock_file
from .sample import Sample
from .transform import Transform
from .working_points import get_working_points

__all__ = [
    "Cuts",
    "Extended_Flavours",
    "Flavours",
    "GitError",
    "Label",
    "LabelContainer",
    "Labeller",
    "Sample",
    "Transform",
    "__version__",
    "calculate_best_fraction_values",
    "check_for_fork",
    "check_for_uncommitted_changes",
    "create_and_push_tag",
    "get_git_hash",
    "get_mock_file",
    "get_working_points",
    "hdf5",
    "is_git_repo",
    "utils",
]



from __future__ import annotations

from typing import TYPE_CHECKING

import h5py
import numpy as np
from numpy.lib.recfunctions import unstructured_to_structured as u2s

from ftag.transform import Transform

if TYPE_CHECKING:  # pragma: no cover
    import h5py

__all__ = [
    "cast_dtype",
    "compare_groups",
    "extract_group_full",
    "get_dtype",
    "join_structured_arrays",
    "structured_from_dict",
    "write_group_full",
]


def compare_groups(g1: h5py.Group | dict, g2: h5py.Group | dict, path: str = ""):
    """Recursively compare two h5py.Groups or in-memory dicts.

    Parameters
    ----------
    g1 : h5py.Group | dict
        First group or dict to compare
    g2 : h5py.Group | dict
        Second group or dict to compare
    path : str, optional
        Path to the current group, by default ""

    Raises
    ------
    TypeError
        If the types of the items do not match
    """
    assert set(g1.keys()) == set(g2.keys()), (
        f"{path}: Keys mismatch: {set(g1.keys())} vs {set(g2.keys())}"
    )

    for key in g1:
        item1 = g1[key]
        item2 = g2[key]
        subpath = f"{path}/{key}"

        # Check for extracted group (dict) with 'data'
        if isinstance(item1, dict) and "data" in item1:
            assert isinstance(item2, dict), f"{subpath}: One side missing 'data'"
            assert "data" in item2, f"{subpath}: One side missing 'data'"
            np.testing.assert_array_equal(
                item1["data"], item2["data"], err_msg=f"{subpath}: Data mismatch"
            )
            assert item1.get("attrs", {}) == item2.get("attrs", {}), f"{subpath}: Attr mismatch"

        # Check for full extracted group
        elif isinstance(item1, dict):
            assert isinstance(item2, dict), f"{subpath}: Expected nested dict"
            compare_groups(item1, item2, subpath)

        # h5py.Dataset or Group objects
        elif isinstance(item1, h5py.Dataset):
            np.testing.assert_array_equal(item1[()], item2[()], err_msg=f"{subpath}: Data mismatch")
            assert dict(item1.attrs) == dict(item2.attrs), f"{subpath}: Attr mismatch"

        elif isinstance(item1, h5py.Group):
            compare_groups(item1, item2, subpath)

        else:
            raise TypeError(f"{subpath}: Unexpected type: {type(item1)}")


def write_group_full(h5group: h5py.Group, data: dict):
    """Write a nested dictionary structure to an HDF5 group.

    This function recursively writes a dictionary containing datasets and subgroups
    to an HDF5 group. The dictionary should have the structure created by
    extract_group_full().

    Parameters
    ----------
    h5group : h5py.Group
        The HDF5 group to write data to
    data : dict
        Dictionary containing the data structure to write. Can contain:
        - '_group_attrs': dict of group-level attributes
        - dataset entries: dict with 'data' and 'attrs' keys
        - subgroup entries: nested dictionaries

    Raises
    ------
    TypeError
        If an unexpected value type is encountered in the data dict
    """
    # Write group-level attributes
    if "_group_attrs" in data:
        for k, v in data["_group_attrs"].items():
            h5group.attrs[k] = v

    for key, value in data.items():
        if key == "_group_attrs":
            continue
        if isinstance(value, dict) and "data" in value:
            dset = h5group.create_dataset(key, data=value["data"])
            for attr_k, attr_v in value["attrs"].items():
                dset.attrs[attr_k] = attr_v
        elif isinstance(value, dict):
            subgroup = h5group.create_group(key)
            write_group_full(subgroup, value)
        else:
            raise TypeError(f"Unexpected value type for key '{key}': {type(value)}")


def extract_group_full(group: h5py.Group) -> dict:
    """Extract the full contents of an HDF5 group into a nested dictionary.

    This function recursively extracts all datasets, subgroups, and attributes
    from an HDF5 group into an in-memory dictionary structure. Group-level
    attributes are stored under the '_group_attrs' key.

    Parameters
    ----------
    group : h5py.Group
        The HDF5 group to extract data from

    Returns
    -------
    dict
        Nested dictionary containing:
        - '_group_attrs': dict of group-level attributes (if any)
        - dataset entries: dict with 'data' (array) and 'attrs' (dict) keys
        - subgroup entries: nested dictionaries with same structure

    Raises
    ------
    TypeError
        If an unsupported HDF5 item type is encountered
    """
    result = {}
    # Save group-level attributes
    if group.attrs:
        result["_group_attrs"] = {k: group.attrs[k] for k in group.attrs}
    for key, item in group.items():
        if isinstance(item, h5py.Dataset):
            result[key] = {"data": item[()], "attrs": {k: item.attrs[k] for k in item.attrs}}
        elif isinstance(item, h5py.Group):
            result[key] = extract_group_full(item)
        else:
            raise TypeError(f"Unsupported item {key}: {type(item)}")
    return result


def get_dtype(
    ds: h5py.Dataset,
    variables: list[str] | None = None,
    precision: str | None = None,
    transform: Transform | None = None,
    full_precision_vars: list[str] | None = None,
) -> np.dtype:
    """Return a dtype based on an existing dataset and requested variables.

    Parameters
    ----------
    ds : h5py.Dataset
        Input h5 dataset
    variables : list[str] | None, optional
        List of variables to include in dtype, by default None
    precision : str | None, optional
        Precision to cast floats to, "half" or "full", by default None
    transform : Transform | None, optional
        Transform to apply to variables names, by default None
    full_precision_vars : list[str] | None, optional
        List of variables to keep in full precision, by default None

    Returns
    -------
    np.dtype
        Output dtype

    Raises
    ------
    ValueError
        If variables are not found in dataset
    """
    variables = variables or ds.dtype.names
    # If we have a non structured array we just return its dtype
    if variables is None:
        return ds.dtype

    if full_precision_vars is None:
        full_precision_vars = []
    if (missing := set(variables) - set(ds.dtype.names)) and transform is not None:
        variables = transform.map_variable_names(ds.name, variables, inverse=True)
        missing = set(variables) - set(ds.dtype.names)
    if missing:
        raise ValueError(
            f"Variables {missing} were not found in dataset {ds.name} in file {ds.file.filename}"
        )

    dtype = [(n, x) for n, x in ds.dtype.descr if n in variables]
    if precision:
        dtype = [
            (n, cast_dtype(x, precision)) if n not in full_precision_vars else (n, x)
            for n, x in dtype
        ]

    return np.dtype(dtype)


def cast_dtype(typestr: str, precision: str) -> np.dtype:
    """Cast float type to half or full precision.

    Parameters
    ----------
    typestr : str
        Input type string
    precision : str
        Precision to cast to, "half" or "full"

    Returns
    -------
    np.dtype
        Output dtype

    Raises
    ------
    ValueError
        If precision is not "half" or "full"
    """
    t = np.dtype(typestr)
    if t.kind != "f":
        return t

    if precision == "half":
        return np.dtype("f2")
    if precision == "full":
        return np.dtype("f4")
    raise ValueError(f"Invalid precision {precision}")


def join_structured_arrays(arrays: list) -> np.ndarray:
    """Join a list of structured numpy arrays.

    See https://github.com/numpy/numpy/issues/7811

    Parameters
    ----------
    arrays : list
        List of structured numpy arrays to join

    Returns
    -------
    np.ndarray
        A merged structured array
    """
    dtype: list = sum((a.dtype.descr for a in arrays), [])
    newrecarray = np.empty(arrays[0].shape, dtype=dtype)
    for a in arrays:
        for name in a.dtype.names:
            newrecarray[name] = a[name]

    return newrecarray


def structured_from_dict(d: dict[str, np.ndarray]) -> np.ndarray:
    """Convert a dict to a structured array.

    Parameters
    ----------
    d : dict[str, np.ndarray]
        Input dict of numpy arrays

    Returns
    -------
    np.ndarray
        Structured array
    """
    arrays = np.column_stack(list(d.values()))
    dtypes = np.dtype([(k, v.dtype) for k, v in d.items()])
    return u2s(arrays, dtype=dtypes)
